from pathlib import Path
from platformdirs import user_cache_dir

import pytest

from redplanet.user_config import (
	get_dirpath_datacache,
	set_dirpath_datacache,
)





## `get_dirpath_datacache` (public function)
class Test__get_dirpath_datacache:

	## Valid input, default value
	def test__get_dirpath_datacache__default_value(self):
		current_path = get_dirpath_datacache()
		print(f'\n\t- Default data cache path: {current_path}')

		assert current_path == Path(user_cache_dir(appname="redplanet")).resolve()





## `set_dirpath_datacache` (public function)
class Test__set_dirpath_datacache:

	## Valid input: str
	def test__set_dirpath_datacache__valid_string(self):
		valid_path_str = "/tmp/redplanet_cache"
		set_dirpath_datacache(valid_path_str)

		assert get_dirpath_datacache() == Path(valid_path_str).resolve()


	## Valid input: pathlib.Path
	def test__set_dirpath_datacache__valid_path_object(self):
		valid_path_obj = Path("/tmp/redplanet_cache")
		set_dirpath_datacache(valid_path_obj)

		assert get_dirpath_datacache() == valid_path_obj.resolve()


	## Invalid input: str not valid path
	def test__set_dirpath_datacache__invalid_string(self):
		invalid_path_str = "\0"

		with pytest.raises(ValueError, match='Invalid path string provided:'):
			set_dirpath_datacache(invalid_path_str)


	## Invalid input: type not str or Path
	def test__set_dirpath_datacache__invalid_type(self):
		invalid_path_type = 12345

		with pytest.raises(TypeError, match='Path must be a string or a Path object.'):
			set_dirpath_datacache(invalid_path_type)



import pytest
import numpy as np

from redplanet import Mag



def test_load_get_valid():
    df = Mag.depth.get_dataset()

    assert df.shape == (412, 6)
    assert df.chi2_reduced.mean() == 1.2787405731700439

    near = Mag.depth.get_nearest(lon=10,lat=10,as_dict=True)[0]

    assert np.allclose(near['depth_km'], np.array([4., 2., 4.]))
    assert np.allclose(near['distance_km'], 283.51013951489193)



from collections.abc import Callable
import re
from textwrap import dedent, indent

from redplanet.helper_functions.docstrings.direct_subs.consts import _dict_fragments


## python multiline strings are weird -- remove the common indentation and first/last newlines.
__cached_fragments = {
    k: dedent(v.rstrip()[1:])
    for k, v in _dict_fragments.items()
}

## match the key (e.g. `{param.lon}`), assuming it appears at the START of a line and might be indented (i.e. leading whitespace)
__pattern = re.compile(r"^( *)\{([^}]+)\}", re.MULTILINE)

## replace the key with the corresponding fragment, indented to match the placeholder’s indent.
def __repl(match):
    current_indent, key = match.groups()
    fragment = __cached_fragments.get(key)
    if fragment is not None:
        ## re-indent the fragment to match the placeholder’s indent.
        return indent(fragment, current_indent)
    ## if the key isn't found, leave the placeholder unchanged.
    return match.group(0)


def _substitute_direct(func: Callable) -> Callable:
    ## replace all occurrences in a single pass over the docstring.
    func.__doc__ = __pattern.sub(__repl, func.__doc__)
    return func



import tempfile
from pathlib import Path
from hashlib import sha256

import pytest
import xxhash

from redplanet.DatasetManager.hash import (
    _hashalgs,
    _calculate_hash_from_file,
    _calculate_hash_from_url,
)





## `_hashalgs` (private variable) -- representing available hash algorithms
class Test__hashalgs:

    ## Verify types/attrs of keys/values
    def test__hashalgs__types(self):
        assert isinstance(_hashalgs, dict)
        for alg, hash_obj in _hashalgs.items():
            assert isinstance(alg, str)
            assert hasattr(hash_obj(), 'update')
            assert hasattr(hash_obj(), 'hexdigest')





## `_calculate_hash_from_file` (internal function) -- calculate file hash with specified algorithm
class Test__calculate_hash_from_file:

    ## Valid input
    def test__calculate_hash_from_file__valid(self, tmp_path):
        # create temp file to hash
        temp_file_path = tmp_path / "testfile"
        temp_file_path.write_bytes(b'Test content for hashing')

        expected_sha256 = sha256(b'Test content for hashing').hexdigest()
        expected_xxh3_64 = xxhash.xxh3_64(b'Test content for hashing').hexdigest()

        assert _calculate_hash_from_file(fpath=temp_file_path, alg='sha256') == expected_sha256
        assert _calculate_hash_from_file(fpath=temp_file_path, alg='xxh3_64') == expected_xxh3_64


    ## Invalid input: `fpath` does not exist
    def test__calculate_hash_from_file__invalid_file(self):
        non_existent_file = Path('present_day___present_time___hahahahaha.txt')

        with pytest.raises(FileNotFoundError):
            _calculate_hash_from_file(fpath=non_existent_file, alg='sha256')


    ## Invalid input: `alg` not available
    def test__calculate_hash_from_file__invalid_alg(self, tmp_path):
        # create temp file to hash
        temp_file_path = tmp_path / "testfile"
        temp_file_path.write_bytes(b'Test content for hashing')

        with pytest.raises(ValueError, match="Unsupported algorithm"):
            _calculate_hash_from_file(fpath=temp_file_path, alg='nonexistent_algorithm')





## `_calculate_hash_from_url` (internal function) -- calculate hash from URL without downloading it, with specified algorithm
class Test__calculate_hash_from_url:

    ## Valid input
    def test__calculate_hash_from_url__valid(self):
        known_url = 'https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html'
        known_sha256 = 'c19249281275b6c8e3e7ba90edbfbde1998c47f4a3e8171c9a9ed46458495e75'
        known_xxh3_64 = '22072ea3f116681d'

        assert _calculate_hash_from_url(url=known_url, alg='sha256') == known_sha256
        assert _calculate_hash_from_url(url=known_url, alg='xxh3_64') == known_xxh3_64


    ## Invalid input: `url` is invalid
    def test__calculate_hash_from_url__invalid_url(self):
        with pytest.raises(ValueError, match="Invalid URL"):
            _calculate_hash_from_url(url='invalid_url', alg='sha256')


    ## Invalid input: `alg` not available
    def test__calculate_hash_from_url__invalid_alg(self):
        with pytest.raises(ValueError, match="Unsupported algorithm"):
            _calculate_hash_from_url(url='https://www.google.com', alg='nonexistent_algorithm')



from tqdm import tqdm

from redplanet import user_config

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.DatasetManager.dataset_info import _DATASETS


def prefetch() -> None:
    """
    Download a few key datasets to cache. This is primarily meant for the demo notebook.
    """

    print(f'Cache directory: {user_config.get_dirpath_datacache()}')

    ## declare which datasets to download (these correspond to keys in `redplanet.DatasetManager.dataset_info._DATASETS`)
    names = [
        'GRS_v2',
        'dichotomy_coords',
        # 'DEM_200m',
        # 'DEM_463m',
        # 'SH_5km',
        'SH_10km',
        'moho_registry',
        'crater_db',
        'Langlais2019',
        # 'Genova2016',
        'MOLA_shape_719',
        'Gong & Weiczorek, 2021',
    ]

    # ## DEBUG: delete all if they exist
    # for name in names:
    #     fpath = _get_fpath_dataset(name)
    #     if fpath.exists():
    #         fpath.unlink()

    ## calculate size of all datasets (not accounting for what's already downloaded)
    sum_size_mib = 0
    for name in names:
        sum_size_mib += _DATASETS[name]['size_mib']
    print(f'(Total download size: {sum_size_mib:.2f} MiB)')
    print()

    # ## download all (superseded by progress bar)
    # for name in names:
    #     print(f'Downloading: "{name}" ({_DATASETS[name]["size_mib"]:0.2f} MiB)')
    #     fpath = _get_fpath_dataset(name)

    ## download all, with progress bar
    with tqdm(total=len(names), desc="Downloading datasets") as pbar:
        for name in names:
            pbar.set_postfix(current=f'"{name}" ({_DATASETS[name]["size_mib"]:0.2f} MiB)')
            _get_fpath_dataset(name)
            pbar.update(1)



import pandas as pd

from redplanet.Craters.loader import _get_dataset

from redplanet.helper_functions.coordinates import _verify_coords
from redplanet.helper_functions.docstrings.main import substitute_docstrings



@substitute_docstrings
def get(
    crater_id : str | list[str]     = None,
    name      : str | list[str]     = None,
    lon       : tuple[float, float] = None,
    lat       : tuple[float, float] = None,
    diameter  : tuple[float, float] = None,
    has_age   : bool                = None,
    as_dict   : bool                = False,
) -> pd.DataFrame | list[dict]:
    """
    Filter/query a dataset of craters >50km diameter, with ages/names when available. Calling this with no arguments will return the full dataset.

    We create a custom dataset (0.28 MiB) which unifies the following:

    1. Global database of Martian impact craters ({@Robbins2012_crater_db.p}).
    2. Crater ages from both Hartmann and Neukum isochron methods ({@Robbins2013_crater_ages.p}).
    3. IAU-approved crater nomenclature ({@IAU_crater_names.p}}).

    For more information and code to reproduce the dataset, see <https://github.com/Humboldt-Penguin/redplanet/tree/main/datasets/Craters>{target="_blank"} -- TODO: I'll eventually have a section on my website to describe datasets and how we modified them, add a link to that here.

    Parameters
    ----------
    crater_id : str | list[str], optional
        Unique crater identifier formatted ##-######, where the first two numbers indicate the Mars subquad and the last six number the craters in that subquad from largest to smallest diameter.
    name : str | list[str], optional
        Crater name according to official IAU nomenclature (as of 2024-11-26).
    lon : tuple[float, float], optional
        Filter craters whose center falls within this range of longitudes.

        The given range must be a subset of either [-180,180] or [0,360] -- e.g. `lon=[-170,350]` is not allowed (it doesn't make sense).
    lat : tuple[float, float], optional
        Filter craters whose center falls within this range of latitudes.
    diameter : tuple[float, float], optional
        Filter craters whose diameter falls within this range, in kilometers.
    has_age : bool, optional
        If True, only return craters with both Hartmann/Neukum isochron ages available. Default is False.
    as_dict : bool, optional
        If True, return the crater dataset as a list of dictionaries instead of a pandas DataFrame. Default is False.

    Returns
    -------
    pd.DataFrame | list[dict]
        Filtered list of craters with columns:

        - `id` : str
            - Unique crater identifier formatted ##-######, where the first two numbers indicate the Mars subquad and the last six number the craters in that subquad from largest to smallest diameter.
        - `name` : str
            - Crater name according to official IAU nomenclature (as of 2024-11-26).
        - `lat` : float
            - Latitude of the crater center.
        - `lon` : float
            - Longitude of the crater center, in range [-180,180].
        - 'plon' : float
            - Longitude of the crater center, in range [0,360].
        - `diam` : float
            - Diameter of the crater, in km.
        - `['diam_sd', 'diam_elli_major', 'diam_elli_minor', 'diam_elli_angle', 'diam_elli_major_sd', 'diam_elli_minor_sd']` : float
            - For more info, see Appendix A of {@Robbins2012_crater_db.n}.
        - `['N_H(10)', 'N_N(10)', 'N_H(25)', 'N_N(25)', 'N_H(50)', 'N_N(50)', 'Hartmann Isochron Age', 'Neukum Isochron Age', 'Hartmann Turn-Off Diameter', 'Neukum Turn-Off Diameter']` : None | list[float, float, float]
            - When available, the ages are given in a list where the first value is the estimated age and second/third are uncertainties (they will always be negative/positive respectively). All values are in billions of years (aka "giga-annums"/"Ga").
            - For more info, see Supplementary Table 3 of {@Robbins2013_crater_ages.n}.
    """

    df = _get_dataset()

    if crater_id:
        if isinstance(crater_id, str):
            crater_id = [crater_id]
        df = df[ df['id'].isin(crater_id) ]

    if name:
        if isinstance(name, str):
            name = [name]
        df = df[ df['name'].isin(name) ]


    if lon:
        _verify_coords(lon, 0)
        # lon = _plon2slon(lon)    ## this introduces unexpected/annoying behavior, TODO figure it out eventually lol (or add a plon col to df, and if any input lons are >180 then query that column instead lol)
        if lon[0] > 180 or lon[1] > 180:
            df = df[ df['plon'].between(lon[0], lon[1]) ]
        else:
            df = df[ df['lon'].between(lon[0], lon[1]) ]

    if lat:
        _verify_coords(0, lat)
        df = df[ df['lat'].between(lat[0], lat[1]) ]


    if diameter:
        df = df[ df['diam'].between(diameter[0], diameter[1]) ]

    if has_age:
        df = df[
            pd.notna(df['Hartmann Isochron Age']) &
            pd.notna(df['Neukum Isochron Age'])
        ]

    if as_dict:
        df = df.to_dict(orient='records')

    return df



import numpy as np

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.GriddedData import GriddedData

from redplanet.helper_functions.docstrings.main import substitute_docstrings





_dat_boug: GriddedData | None = None

@substitute_docstrings
def get_dataset() -> GriddedData:
    """
    {fulldoc.get_dataset_GriddedData}
    """
    if _dat_boug is None:
        raise ValueError('Bouguer dataset not loaded. Use `redplanet.Crust.boug.load(<model_params>)`.')
    return _dat_boug

@substitute_docstrings
def get_metadata() -> dict:
    """
    {fulldoc.get_metadata}
    """
    return dict(get_dataset().metadata)





@substitute_docstrings
def load(model: str = None) -> None:
    """
    Load Bouguer gravity anomaly dataset.

    Parameters
    ----------
    model : str
        Name of the Bouguer model to load. Options are:

        - `'Genova2016'` (127 MiB) — Bouguer gravity anomaly map from {@Genova2016_boug_data.n}, computed from truncated GMM-3 solution (degree 2 to 90) ({@Genova2016_boug_paper.p}).

    Raises
    ------
    ValueError
        If an invalid model name is provided.
    """

    ## I expect to add more later, so users should explicitly choose Genova2016 for forward compatibility.
    info = {
        'Genova2016': {
            'shape'          : (2880, 5760),
            'dtype'          : np.float64,
            'post-processing': lambda arr: np.flipud(arr),
            'lon'            : np.linspace(  0, 360, 5760, endpoint=False) + 1 / (2 * 16),
            'lat'            : np.linspace(-90,  90, 2880, endpoint=False) + 1 / (2 * 16),
            'metadata': {
                'description': 'Bouguer gravity anomaly map computed from truncated GMM-3 solution (degree 2 to 90).',
                'units'      : 'mGal',
                'links'      : {
                    'download': 'https://pds-geosciences.wustl.edu/mro/mro-m-rss-5-sdp-v1/mrors_1xxx/data/rsdmap/',
                    'lbl'     : 'https://pds-geosciences.wustl.edu/mro/mro-m-rss-5-sdp-v1/mrors_1xxx/data/rsdmap/ggmro_120_bouguer_90.lbl',
                },
            },
        },
    }

    if model not in info:
        raise ValueError(f"Invalid Bouguer model: '{model}'. Options are: {list(info.keys())}.")



    if model == 'Genova2016':

        fpath = _get_fpath_dataset(model)

        dat = np.memmap(
            fpath,
            mode  = 'r',
            dtype = info[model]['dtype'],
            shape = info[model]['shape'],
        )
        dat = info[model]['post-processing'] (dat)

        metadata = info[model]['metadata']
        metadata['fpath'] = fpath

        global _dat_boug
        _dat_boug = GriddedData(
            lon       = info[model]['lon'],
            lat       = info[model]['lat'],
            is_slon   = False,
            data_dict = {'boug': dat},
            metadata  = metadata,
        )



    else:
        raise ValueError(f"THE DEVELOPER MESSED UP. THIS SHOULD NOT HAPPEN.")

    return



from pathlib import Path
from platformdirs import user_cache_dir
import time

import pytest

from redplanet.user_config import get_dirpath_datacache, set_dirpath_datacache
from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.DatasetManager.dataset_info import _get_download_info
from redplanet.DatasetManager.hash import _calculate_hash_from_file

## `_get_fpath_dataset` (internal function) -- download/cache dataset
class Test__get_fpath_dataset:

    ## Valid input: default & custom cache dirs
    def test__get_fpath_dataset__valid_default_cachedir(self, tmp_path):

        base_dirs = [
            Path(Path(user_cache_dir(appname='redplanet')).resolve()),
            tmp_path,   ## TODO: p sure there's a better way to test multiple inputs but idk if it works with my fixtures/logic, too lazy rn
        ]

        for base_dir in base_dirs:

            if base_dir == tmp_path: set_dirpath_datacache(base_dir)

            ## use an actual dataset instead of mock file so we can test full integration -- mock file test in `test_download.py`
            fpath_expected: Path =  base_dir / 'Crust' / 'dichotomy' / 'dichotomy_coordinates-JAH-0-360.txt'
            known_hash: str = _get_download_info('dichotomy_coords')['hash']['sha256']

            ## download file to cache & verify
            if (fpath_expected.is_file()): fpath_expected.unlink()
            fpath_actual = _get_fpath_dataset('dichotomy_coords')

            assert fpath_actual.is_file()
            assert fpath_actual == fpath_expected
            assert _calculate_hash_from_file(fpath_actual, 'sha256') == known_hash

            ## now try to download again, but recognize that the file already exists in cache since path returns immediately
            t0 = time.time()
            fpath_actual = _get_fpath_dataset('dichotomy_coords')
            assert time.time() - t0 < 0.1, "Dataset was previously downloaded to cache, but upon second access, we couldn't find it."

            ## cleanup (TODO: maybe this should be in a fixture...? I don't always wanna delete it though, maybe make a mock dataset file instead [note this contradicts previous comment about why using actual dataset instead of mock file for full integration])
            fpath_actual.unlink()



import pandas as pd

from redplanet.DatasetManager.main import _get_fpath_dataset



def get_registry() -> pd.DataFrame:
    """
    Get a list of all available Moho models.

    Returns
    -------
    pd.DataFrame
        DataFrame with the following columns: ['interior_model', 'insight_thickness', 'rho_south', 'rho_north']. For an explanation of these columns, see parameters of `redplanet.Crust.moho.load()`.
    """
    registry = pd.read_csv(
        _get_fpath_dataset('moho_registry'),
        usecols = ['model_name'],
    )
    registry = registry['model_name'].str.split('-', expand=True)
    registry.columns = ['interior_model', 'insight_thickness', 'rho_south', 'rho_north']
    registry.iloc[:,1:] = registry.iloc[:,1:].astype(int)
    return registry



'''
For info/sources on reference interior models, see:
    1. https://markwieczorek.github.io/ctplanet/source/generated/ctplanet.ReadRefModel.html
    2. https://github.com/MarkWieczorek/ctplanet/tree/74e8550080d4adc68ae291a500e8d198a40d437c/examples/Data/Mars-reference-interior-models
'''
_interior_models: list[str] = (
    'DWAK',
    'DWThot',
    'DWThotCrust1',
    'DWThotCrust1r',
    'EH45Tcold',
    'EH45TcoldCrust1',
    'EH45TcoldCrust1r',
    'EH45ThotCrust2',
    'EH45ThotCrust2r',
    'Khan2022',
    'LFAK',
    'SANAK',
    'TAYAK',
    'YOTHotRc1760kmDc40km',
    'YOTHotRc1810kmDc40km',
    'ZG_DW',
)



import pytest

import redplanet.user_config as rpcfg

@pytest.fixture(autouse=True)
def reset_user_config():
    """
    Reset values in `redplanet.user_config` before every test.
    """
    rpcfg.datacache._dirpath_datacache                  = None
    rpcfg.max_hash_size._max_size_to_calculate_hash_GiB = None
    rpcfg.stream_hash_check._enable_stream_hash_check   = True



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "pandas",
# ]
# ///

from pathlib import Path

import pandas as pd



def main():
    fpath_registry = Path('/home/lain/root/100_work/110_projects/111_mars/code-repos/redplanet_datasets__cloud_upload_V2wUV/output/redplanet_registry_2024-10-19_18-35.csv')
    reg_old = pd.read_csv(fpath_registry).to_dict(orient='list')

    reg_new = []

    needle = 'Moho-Mars-'

    for filepath, sha1, url in zip(reg_old['filepath'], reg_old['sha1'], reg_old['download_url']):
        filestem = Path(filepath).stem
        if not filestem.startswith(needle):
            continue
        model_name = filestem[ len(needle) : ]
        download_code = url.split('/')[-1][:-3]
        reg_new.append([model_name, download_code, sha1])

    reg_new = pd.DataFrame(reg_new, columns=['model_name', 'box_download_code', 'sha1'])
    reg_new.sort_values(by='model_name', inplace=True)

    reg_new.to_csv('moho_registry.csv', index=False)



if __name__ == '__main__':
    main()



'''
Basic search function:

    ```
    import pandas as pd

    def search_moho_registry(model_name, fpath_moho_registry) -> list[str,str]:
        """
        Parameters:
            - `model_name`: str
                - Model name in the format 'MODEL-THICK-RHOS-RHON', e.g. 'Khan2022-38-2900-2900'.

        Returns:
            - list[str,str]
                - `box_download_code`, which yields a download link when appended to 'https://rutgers.box.com/shared/static/'.
                - `sha1` hash.
        """
        df = pd.read_csv(fpath_moho_registry)
        result = df[ df['model_name'] == model_name ]
        if result.empty:
            result = None
        else:
            result = result.values.tolist()[0][1:]
        return result


    model_name = 'Khan2022-38-2900-2900'
    fpath_moho_registry = Path('moho_registry.csv')
    result = search_moho_registry(model_name, fpath_moho_registry)
    print(f'{result = }')
    ```








------------------------------


ALT REGISTRY METHOD (nested json which gets rid of redundant text by structuring/aceessing like `registry['Khan2022'][38][2900][2900]`, but it ends up being the same size ~2mb as a plain csv lol rip):

    ```
    from pathlib import Path
    import json

    import pandas as pd



    fpath_registry = Path('/home/lain/root/100_work/110_projects/111_mars/code-repos/redplanet_datasets__cloud_upload_V2wUV/output/redplanet_registry_2024-10-19_18-35.csv')

    registry = pd.read_csv(fpath_registry)
    registry = registry.to_dict(orient='list')

    new = {}

    for filepath, sha1, url in zip(registry['filepath'], registry['sha1'], registry['download_url']):
        if 'Moho-Mars' not in filepath:
            continue

        url = url.split('/')[-1]

        filename = Path(filepath).stem
        moho_code = filename.split('-')[2:]

        for i in [1,2,3]:
            moho_code[i] = int(moho_code[i])

        pointer = new
        for value in moho_code[:-1]:
            if value not in pointer:
                pointer[value] = {}
            pointer = pointer[value]
        pointer[ moho_code[-1] ] = [url, sha1]



    fpath_new_registry = Path('registry.json')
    if fpath_new_registry.is_file(): fpath_new_registry.unlink()

    with open(fpath_new_registry, 'w') as f:
        json.dump(new, f)



    ######################################################
    """ensure new registry was constructed properly"""

    i = 0
    hashes = set()

    def count_items(d):
        global i
        global hashes
        for value in d.values():
            if isinstance(value, dict):
                count_items(value)
            else:
                i += 1
                this_hash = value[1]
                # if this_hash in hashes:
                #     print(f'{value = }')
                hashes.add(this_hash)

    """
    Expected output:
    num entries: 21894
    num unique files (by hash): 17092
    """

    count_items(new)
    print(f'num entries: {i}')
    print(f'num unique files (by hash): {len(hashes)}')
    ```



'''



import numpy as np


class CoordinateError(Exception):
    """
    Exception raised for coordinate values outside of acceptable range.
    """
    pass


def _verify_coords(
    lon: float | list | np.ndarray,
    lat: float | list | np.ndarray,
) -> None:
    """
    Verify that longitude and latitude values are within valid ranges ([-180, 360] and [-90, 90] respectively), otherwise raise a CoordinateError.

    Parameters
    ----------
    lon : float | list | np.ndarray
        Longitude value(s) to validate. Values are expected to be in the range [-180, 360].
    lat : float | list | np.ndarray
        Latitude value(s) to validate. Values are expected to be in the range [-90, 90].

    Raises
    ------
    CoordinateError
        If any longitude value is outside the range [-180, 360] or if any latitude value is outside the range [-90, 90].
    """
    lon = np.atleast_1d(lon)
    lat = np.atleast_1d(lat)

    invalid_lon = np.where((lon < -180) | (lon > 360))[0]
    invalid_lat = np.where((lat < -90) | (lat > 90))[0]

    if invalid_lon.size > 0:
        invalid_lon_values = lon[invalid_lon]
        error_msg = [
            f"Longitude coordinates must be in range [-180, 360].",
            f"The following input values were outside of this range:\n\t- {invalid_lon_values.tolist()}.",
            f"Corresponding input array indices:\n\t- {invalid_lon.tolist()}.",
        ]
        raise CoordinateError("\n".join(error_msg))

    if invalid_lat.size > 0:
        invalid_lat_values = lat[invalid_lat]
        error_msg = [
            f"Latitude coordinates must be in range [-90, 90].",
            f"The following input values were outside of this range:\n\t- {invalid_lat_values.tolist()}.",
            f"Corresponding input array indices:\n\t- {invalid_lat.tolist()}.",
        ]
        raise CoordinateError("\n".join(error_msg))



def _plon2slon(
    plon: float | list | np.ndarray
) -> float | list | np.ndarray:
    """
    Convert positive longitudes (in range [0, 360]) to signed longitudes (in range [-180, 180]). Inputs in range [-180, 0] are returned as is.

    Parameters
    ----------
    plon : float | list | np.ndarray
        Longitude value(s) in positive format.

    Returns
    -------
    float | list | np.ndarray
        Converted longitude value(s) in signed format. The return type matches the input type.

    Notes
    -----
    Actual mapping over the full input range:

    - [-180, 180) --> [-180, 180)
    - [180, 360]  --> [-180, 0]

    Self reminder:

    - Signed longitude   [-180, 180]  -->  Arabia Terra in middle & Hellas on right.
    - Positive longitude [0, 360]     -->  Olympus Mons in middle & Hellas on left.
    """
    def convert(plon):
        return ((plon - 180) % 360) - 180

    if isinstance(plon, (float, np.ndarray)):
        return convert(plon)
    else:
        return convert(np.array(plon)).tolist()


def _slon2plon(
    slon: float | list | np.ndarray
) -> float | list | np.ndarray:
    """
    Convert signed longitudes (in range [-180, 180]) to positive longitudes (in range [0, 360]). Inputs in range [180, 360] are returned as is.

    Parameters
    ----------
    slon : float | list | np.ndarray
        Longitude value(s) in signed format.

    Returns
    -------
    float | list | np.ndarray
        Converted longitude value(s) in positive format. The return type matches the input type.

    See Also
    --------
    _plon2slon : Converts positive longitudes to signed longitudes.
    """
    def convert(slon):
        return slon % 360

    if isinstance(slon, (float, np.ndarray)):
        return convert(slon)
    else:
        return convert(np.array(slon)).tolist()



from pathlib import Path

import pandas as pd


_DATASETS = {
    # 'GRS': {
    #     'url'    : 'https://rutgers.box.com/shared/static/3u8cokpvnbpl8k7uuka7qtz1atj9pxu5',
    #     'fname'  : '2022_Mars_Odyssey_GRS_Element_Concentration_Maps.zip',
    #     'dirpath': 'GRS/',
    #     'hash'   : {
    #         'sha256': 'ba2b5cc62b18302b1da0c111101d0d2318e69421877c4f9c145116b41502777b',
    #     },
    # },
    'GRS_v2': {  # TODO: rename this to 'GRS' at some point
        'url'    : 'https://data.mendeley.com/public-files/datasets/3jd9whd78m/files/a0c6ddc8-fbca-4119-b6f5-533a5775b719/file_downloaded',
        # 'url'    : 'https://rutgers.box.com/shared/static/qxfzuv7ka5c57k5awj83ueknr27dmyi4', # alt mirror on my own box account
        'fname'  : 'Rani2022_new_GRS_data.xlsx',
        'dirpath': 'GRS/',
        'hash'   : {
            'sha256': '06f7b8cfc89a803c714a7dd69beba0d65ba9f26ccf8eb5ddc276e280c4e5b917',
        },
        'size_mib': 0.4064617156982422,
    },
    'dichotomy_coords': {
        'url'    : 'https://rutgers.box.com/shared/static/tekd1w26h9mvfnyw8bpy4ko4v48931ri',
        'fname'  : 'dichotomy_coordinates-JAH-0-360.txt',
        'dirpath': 'Crust/dichotomy/',
        'hash'   : {
            'sha256': '42f2b9f32c9e9100ef4a9977171a54654c3bf25602555945405a93ca45ac6bb2',
        },
        'size_mib': 0.04616832733154297,
    },
    'DEM_200m': {
        'url'    : 'https://rutgers.box.com/shared/static/8xfyuf8qw6sbyqgzjjx0g2twehui9595',
        'fname'  : 'Mars_HRSC_MOLA_BlendDEM_Global_200mp_v2.memmap',
        'dirpath': 'Crust/topo/',
        'hash'   : {
            'xxh3_64': 'e8cc649a36ea4fab',
            'md5'    : '74cedd82aaf200b62ebb64affffe0e7e',
            'sha1'   : '0c7704155a3e9fb6bef284980fdb37aa559457c5',
            'sha256' : '691b6ce6a1cacc5fcea4b95ef1832fac50e421e1ec8f7fb33e5c791396aa4a4f',
        },
        'size_mib': 10856.2561378479,
    },
    'DEM_463m': {
        'url'    : 'https://rutgers.box.com/shared/static/s2iu6egz4og7ht310ctebgrloinf81dq',
        'fname'  : 'Mars_MGS_MOLA_DEM_mosaic_global_463m.memmap',
        'dirpath': 'Crust/topo/',
        'hash'   : {
            'xxh3_64': '6eed1a19495d736f',
            'md5'    : '0f2378e55a01c217b2662b7ba07a3f27',
            'sha1'   : 'f3547e5423bd447179e5126e37b262e4136adcac',
            'sha256' : '7788fa9287c633456fbf2de8b0e674a7e375014d2b58731b45f991be284879c4',
        },
        'size_mib': 2025.1318378448486,
    },
    'SH_5km': {
        'url'    : 'https://rutgers.box.com/shared/static/ff1tmj9ovgqugl524az8zt920nwktlh7',
        'fname'  : 'Mars_MOLA_topography_5km.memmap',
        'dirpath': 'Crust/topo/',
        'hash'   : {
            'xxh3_64': 'c342253082d163dd',
        },
        'size_mib': 69.51227760314941,
    },
    'SH_10km': {
        'url'    : 'https://rutgers.box.com/shared/static/mt2aysgkf4s8vqrdo77fjzsnuz1fmapw',
        'fname'  : 'Mars_MOLA_topography_10km.memmap',
        'dirpath': 'Crust/topo/',
        'hash'   : {
            'xxh3_64': '3a8d69acce1a0799',
        },
        'size_mib': 17.38417625427246,
    },
    'moho_registry': {
        'url'    : 'https://rutgers.box.com/shared/static/dcyysy7k1jbhkzt20hgkyt9qxvij79wn',
        'fname'  : 'moho_registry.csv',
        'dirpath': 'Crust/moho/',
        'hash'   : {
            'sha256': '0be4a1ff14df2ee552034487e91ae358dd2e8a907bc37123bbfa5235d1f98dba',
        },
        'size_mib': 2.0612106323242188,
    },
    'crater_db': {
        'url'    : 'https://rutgers.box.com/shared/static/imou5t0f2inyvx8ldcxfk234oeit7z3q',
        'fname'  : 'craters_with_names_and_ages_50km.csv',
        'dirpath': 'Craters/',
        'hash'   : {
            'xxh3_64': 'ea14d77f25f090c4',
            'md5'    : '4e63fd2a7f1367d131ee606edcdfb5f7',
            'sha1'   : '79113d236836e1d8bb53e517ab3cfc4afad2cac2',
            'sha256' : 'e48808ef670e39e812149e4731634d59964b7b3465b1be38eda920f890125bdc',
        },
        'size_mib': 0.2772712707519531,
    },
    'Langlais2019': {
        'url'    : 'https://zenodo.org/records/3876714/files/Langlais2019.sh.gz?download=1',
        'fname'  : 'Langlais2019.sh.gz',
        'dirpath': 'Mag/sh/',
        'hash'   : {
            'md5'    : '39245feef66555366108aecb6d5c0f67',
            'sha256' : '3cad9e268f0673be1702f1df504a4cbcb8dba4480c7b3f629921911488fe247b',
        },
        'size_mib': 0.08092212677001953,
    },
    'Genova2016': {
        ## directly downloaded from here (also see associated lbl file): https://pds-geosciences.wustl.edu/mro/mro-m-rss-5-sdp-v1/mrors_1xxx/data/rsdmap/
        'url'    : 'https://rutgers.box.com/shared/static/cw5ijchyxi021iwmj6qyw2i07hyt24wq',
        'fname'  : 'ggmro_120_bouguer_90.img',
        'dirpath': 'Crust/boug/',
        'hash'   : {
            'md5'    : '95423874e702b8a55f3a4b17b3ef744a',
            'sha256' : 'd3657c9bd80bf5e452ad0d0216078e9295065b5d1e0a04d0fab7bf22e3b50438',
        },
        'size_mib': 126.5625,
    },
    'MOLA_shape_719': {
        'url'    : 'https://zenodo.org/records/10820719/files/Mars_MOLA_shape_719.bshc.gz?download=1',
        'fname'  : 'Mars_MOLA_shape_719.bshc.gz',
        'dirpath': 'Crust/moho/shape/',
        'hash'   : {
            'md5'    : '77c213978e7403c722e38b78e0202d7c',
        },
        'size_mib': 3.7924137115478516,
    },
    'Gong & Weiczorek, 2021': {
        'url'    : 'https://zenodo.org/records/4686358/files/MarsMagnetizationDepth.zip?download=1',
        'fname'  : 'MarsMagnetizationDepth.zip',
        'dirpath': 'Mag/depth/',
        'hash'   : {
            'md5'    : '16780170ee3e0dccaaf719fc201e4281',
        },
        'size_mib': 0.03770160675048828,
    },
}



def peek_datasets() -> dict:
    """
    Returns a dictionary of all available datasets -- intended for debugging/exploration purposes, should NOT be called in production code.
    """
    return _DATASETS


def _get_download_info(name: str) -> dict:
    """
    Returns information to download a dataset as a dictionary with keys 'url', 'fname', 'dirpath' (relative to data cache directory), and 'hash'.
    """
    info = _DATASETS.get(name)

    if info is None:
        error_msg = [
            f"Dataset not found: '{name}'. Options are: {', '.join(_DATASETS.keys())}",
            f"To see all information about the datasets, run `from redplanet.DatasetManager.dataset_info import _DATASETS; print(_DATASETS)`.",
        ]
        raise DatasetNotFoundError('\n'.join(error_msg))

    return info


def _get_download_info_moho(
    model_name: str,
    fpath_moho_registry: Path,
) -> dict:
    """
    Parameters:
        - `model_name`: str
            - Model name in the format 'MODEL-THICK-RHOS-RHON', e.g. 'Khan2022-38-2900-2900'.
        - `fpath_moho_registry`: Path
            - Path to the CSV file containing the registry of Moho models.
    """

    df = pd.read_csv(fpath_moho_registry)
    result = df[ df['model_name'] == model_name ]

    if result.empty:
        interior_model, insight_thickness, rho_south, rho_north = model_name.split('-')
        raise MohoDatasetNotFoundError(
            f'Moho model not found for the given parameters: {interior_model=}, {insight_thickness=}, {rho_south=}, {rho_north=}\n'
            f'To see available models: `redplanet.Crust.moho.get_registry()`'
        )

    box_download_code, sha1 = result.values.tolist()[0][1:]
    result = {
        'url'    : f'https://rutgers.box.com/shared/static/{box_download_code}',
        'fname'  : f'Moho-Mars-{model_name}.sh',
        'dirpath': 'Crust/moho/shcoeffs/',
        'hash'   : {
            'sha1': sha1,
        },
    }
    return result



class DatasetNotFoundError(Exception):
    pass

class MohoDatasetNotFoundError(Exception):
    pass



import pytest
import numpy as np

from redplanet.helper_functions.coordinates import (
    _plon2slon,
    # _slon2plon,        ## TODO: test this, low priority though
    # _verify_coords,    ## TODO: test this, low priority though
)

class Test__plon2slon:

    test_values = (
        # (input, output)
        (-180  , -180  ),
        (-170  , -170  ),
        (-170.1, -170.1),
        (   0  ,    0  ),
        ( 170  ,  170  ),
        ( 170.1,  170.1),
        ( 180  , -180  ),
        ( 190  , -170  ),
        ( 190.1, -169.9),
        ( 350  ,  -10  ),
        ( 350.1,   -9.9),
        ( 360  ,    0  ),
    )

    def test__scalar(self):
        for x, y in self.test_values:
            yy = _plon2slon(x)
            np.testing.assert_allclose(y, yy)
            assert isinstance( yy, (int,float) )

    def test__list(self):
        x_list, y_list = zip(*self.test_values)
        yy_list = _plon2slon(x_list)
        np.testing.assert_allclose(y_list, yy_list)
        assert isinstance( yy_list, (list,tuple) )

    def test__numpy_array(self):
        x_list, y_list = zip(*self.test_values)
        x_arr = np.array(x_list)
        y_arr = np.array(y_list)
        yy_arr = _plon2slon(x_arr)
        np.testing.assert_allclose(y_arr, yy_arr)
        assert isinstance( yy_arr, np.ndarray )

    def test__empty(self):
        assert _plon2slon([]) == []
        np.testing.assert_allclose( _plon2slon(np.array([])), np.array([]) )



import hashlib
from pathlib import Path
from collections.abc import Callable
import requests

import xxhash



_hashalgs: dict[str, Callable] = {
    'xxh3_64': xxhash.xxh3_64,
    'md5'    : hashlib.md5,
    'sha1'   : hashlib.sha1,
    'sha256' : hashlib.sha256,
}
## Note: There's no significant difference between algorithms for time it takes to hash small files (for <1MB, it's on the order of thousandths of a second) -- but it DOES matter for large files (e.g. a 5GB file, sha256 takes 13 seconds, while xxh3 takes 2 seconds).


def get_available_algorithms() -> list[str]:
    """
    Get a list of supported hashing algorithms.
    """
    return list(_hashalgs.keys())



def _calculate_hash_from_file(fpath: Path, alg: str) -> str:
    """
    Calculate the hash of a file using the specified algorithm.

    Args:
        fpath: Path
            Path to the file.
        alg: str
            Hashing algorithm to use (for options, call `from redplanet.DatasetManager.hash import get_available_algorithms(); print(get_available_algorithms())`).

    Returns:
        str: The hexadecimal hash of the file.
    """

    ## Input validation
    fpath = fpath.resolve()
    if not fpath.is_file():
        raise FileNotFoundError(f"File not found: {fpath}")
    if alg not in _hashalgs:
        raise ValueError(f"Unsupported algorithm: '{alg}'. Options are: {', '.join(get_available_algorithms())}")

    ## Calculate hash
    hash_obj = _hashalgs[alg]()

    with fpath.open('rb') as f:
        CHUNK_SIZE = 2**13    # 2^13=8192 bytes per chunk
        while chunk := f.read(CHUNK_SIZE):
            hash_obj.update(chunk)

    return hash_obj.hexdigest()



def _calculate_hash_from_url(url: str, alg: str) -> str:
    """
    Calculate the hash of a file on the internet given its download URL, without actually downloading it.
    This is intended for verifying the integrity of a file.

    Args:
        url: str
            URL of the file.
        alg: str
            Hashing algorithm to use (for options, call `from redplanet.DatasetManager.hash import get_available_algorithms(); print(get_available_algorithms())`).

    Returns:
        str: The hexadecimal hash of the file.
    """

    ## Input validation
    if alg not in _hashalgs:
        raise ValueError(f"Unsupported algorithm: '{alg}'. Options are: {', '.join(get_available_algorithms())}")

    hash_obj = _hashalgs[alg]()

    ## Make the request with streaming enabled to avoid loading the whole file in memory
    response = requests.get(url, stream=True)

    ## Check if the request was successful
    if response.status_code != 200:
        raise ValueError(f"Failed to fetch the file from URL: {url}. Status code: {response.status_code}")

    ## Read the content in chunks and update the hash object
    CHUNK_SIZE = 2**13  # 8192 bytes per chunk
    for chunk in response.iter_content(CHUNK_SIZE):
        if chunk:  # filter out keep-alive new chunks
            hash_obj.update(chunk)

    return hash_obj.hexdigest()



from pathlib import Path
import requests
from requests.exceptions import RequestException
from time import sleep

def _download_file_from_url(
    url       : str,
    dest_path : Path,
    retries   : int = 3,
    timeout   : int = 10,
    chunk_size: int = 512 * 1024,  # 512KB
) -> None:
    """
    Download a file from a URL to a specified local path.

    Parameters
    ----------
    url : str
        The URL of the file to download.
    dest_path : Path
        The local file path where the downloaded file will be saved.
    retries : int, optional
        The number of times to retry the download in case of failures.
    timeout : int, optional
        The timeout in seconds for the download request.
    chunk_size : int, optional
        The size of each chunk to read from the response.

    Raises
    ------
    RequestException
        If the download fails after the specified number of retries.
    """
    dest_path.parent.mkdir(parents=True, exist_ok=True)

    attempt = 0
    while attempt < retries:
        try:
            with requests.get(url, stream=True, timeout=timeout) as response:
                response.raise_for_status()
                with open(dest_path, 'wb') as f:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:  # Filter out keep-alive new chunks
                            f.write(chunk)
                break  # Exit the retry loop if download succeeds
        except RequestException as e:
            attempt += 1
            if attempt == retries:
                raise e  # Re-raise the exception if max retries exceeded
            sleep(1)  # Wait before retrying



import numpy as np
import pyshtools as pysh

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.GriddedData import GriddedData

from redplanet.helper_functions.docstrings.main import substitute_docstrings

from redplanet.DatasetManager.dataset_info import MohoDatasetNotFoundError
from redplanet.Crust.moho.consts import _interior_models





_dat_moho: GriddedData | None = None

@substitute_docstrings
def get_dataset() -> GriddedData:
    """
    {fulldoc.get_dataset_GriddedData}
    """
    if _dat_moho is None:
        raise ValueError('Bouguer dataset not loaded. Use `redplanet.Crust.moho.load(<model_params>)`.')
    return _dat_moho

@substitute_docstrings
def get_metadata() -> dict:
    """
    {fulldoc.get_metadata}
    """
    return dict(get_dataset().metadata)





@substitute_docstrings
def load(
    interior_model    : str,
    insight_thickness : int | str,
    rho_south         : int | str,
    rho_north         : int | str,
    fail_silently     : bool = False,    ##  False [default] -> return None,  True -> return type(bool)
) -> None | bool:
    """
    Load a model of the Mohorovičić discontinuity (crust-mantle interface) with the given parameters. For a list of valid combinations of parameters (total 21,894 options), run `redplanet.Crust.moho.get_registry()` (returns a pandas DataFrame).

    Spherical harmonic coefficients are precomputed in {@Wieczorek2022_icta.n} and downloaded on-the-fly from our own mirror. The full paper discusses/analyzes the models in detail ({@Wieczorek2022_icta_paper.p}). We process spherical harmonic coefficients with `pyshtools` ({@shtools_code.p}; {@shtools_paper.p}).


    Parameters
    ----------
    interior_model : str
        Name of the interior model used for the mantle and core (see notes for more information).

        Options are: ['DWAK', 'DWThot', 'DWThotCrust1', 'DWThotCrust1r', 'EH45Tcold', 'EH45TcoldCrust1', 'EH45TcoldCrust1r', 'EH45ThotCrust2', 'EH45ThotCrust2r', 'Khan2022', 'LFAK', 'SANAK', 'TAYAK', 'YOTHotRc1760kmDc40km', 'YOTHotRc1810kmDc40km', 'ZG_DW']. See notes for more information.
    insight_thickness : int | str
        Assumed crustal thickness beneath the InSight landing site, in km.
    rho_south : int | str
        Average crustal thickness south of the dichotomy boundary, in kg/m^3.
    rho_north : int | str
        Average crustal thickness north of the dichotomy boundary, in kg/m^3.
    fail_silently : bool, optional
        - If False (default), raise `MohoDatasetNotFoundError` if there is no dataset for the given parameters.
        - If True, return True if the dataset was loaded successfully, or False if there is no dataset for the given parameters.


    Returns
    -------
    None | bool
        Return is determined as follows:

        |                         | **Successfully Loaded** | **Model Doesn't Exist**          |
        | ----------------------- | ----------------------- | -------------------------------- |
        | **fail_silently=False** | `None`                  | Raise `MohoDatasetNotFoundError` |
        | **fail_silently=True**  | `True`                  | `False`                          |


    Raises
    ------
    ValueError
        If `interior_model` is not one of the available options.
    MohoDatasetNotFoundError
        If the dataset is not found and `fail_silently` is False.


    Notes
    -----
    - More information on reference interior models:
        - [`ctplanet` function to read reference interior models (we don't use it, but it may be helpful if you're implementing this yourself)](https://markwieczorek.github.io/ctplanet/source/generated/ctplanet.ReadRefModel.html){target="_blank"}
        - [interior model files](https://github.com/MarkWieczorek/ctplanet/tree/74e8550080d4adc68ae291a500e8d198a40d437c/examples/Data/Mars-reference-interior-models){target="_blank"}
    """

    ## load moho

    if interior_model not in _interior_models:
        raise ValueError(
            f'Unknown interior model: "{interior_model}".\n'
            f'Options are: {", ".join(_interior_models)}.'
        )

    try:
        fpath_moho = _get_fpath_dataset(
            f'Moho-Mars-{interior_model}-{insight_thickness}-{rho_south}-{rho_north}'
        )
    except MohoDatasetNotFoundError as e:
        if fail_silently:
            return False
        else:
            raise

    ds_moho = (
        pysh.SHCoeffs.from_file(fpath_moho)
        .expand()
        .to_xarray()
        .isel(lat=slice(None, None, -1))  ## in pysh, lats are always decreasing at first
    )



    ## load shape

    fpath_shape = _get_fpath_dataset('MOLA_shape_719')

    ds_shape = (
        pysh.SHCoeffs.from_file(
            fpath_shape,
            lmax   = 90,
            format = 'bshc'
        )
        .expand()
        .to_xarray()
        .isel(lat=slice(None, None, -1))
    )



    ## GriddedData

    global _dat_moho

    _dat_moho = GriddedData(
        lon       = ds_moho.lon.values,
        lat       = ds_moho.lat.values,
        is_slon   = False,
        data_dict = {
            'moho'   : ds_moho.values,
            'crthick': (ds_shape - ds_moho).values,
        },
        metadata  = {
            'title': f'{interior_model}-{insight_thickness}-{rho_south}-{rho_north}',
            'units': 'm',
            'model_params': {
                'interior_model'      : interior_model,
                'insight_thickness_km': insight_thickness,
                'rho_south'           : rho_south,
                'rho_north'           : rho_north,
            },
            'lmax': 90,
            'source' : 'https://doi.org/10.5281/zenodo.6477509',
            'fpath': fpath_moho,
        },
    )



    if fail_silently:
        return True
    return



import numpy as np
import pyshtools as pysh

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.GriddedData import GriddedData

from redplanet.helper_functions.docstrings.main import substitute_docstrings





_dat_mag: GriddedData | None = None

@substitute_docstrings
def get_dataset() -> GriddedData:
    """
    {fulldoc.get_dataset_GriddedData}
    """
    if _dat_mag is None:
        raise ValueError('Bouguer dataset not loaded. Use `redplanet.Mag.sh.load(<model_params>)`.')
    return _dat_mag

@substitute_docstrings
def get_metadata() -> dict:
    """
    {fulldoc.get_metadata}
    """
    return dict(get_dataset().metadata)





@substitute_docstrings
def load(
    model : str = None,
    lmax  : int = 134,
) -> None:
    """
    Load a magnetic field model for Mars.

    Parameters
    ----------
    model : str
        Name of the magnetic field model to load. Options are:

        - `'Langlais2019'` (0.081 MiB) — Spherical harmonic model of the magnetic field of Mars with a spatial resolution of ~160 km at the surface, corresponding to spherical harmonic degree 134. Integrates data from MGS magnetometer, MGS electron reflectometer, and MAVEN magnetometer.
            - Dataset downloaded from {@Langlais2019_data.n}. The full paper discusses/analyzes the models in detail ({@Langlais2019_paper.p}).
            - We process spherical harmonic coefficients with `pyshtools` ({@shtools_code.p}; {@shtools_paper.p}).
    lmax : int, optional
        The maximum spherical harmonic degree of the coefficients to load. Default is 134 (maximum for 'Langlais2019').

    Raises
    ------
    ValueError
        If an invalid model name is provided.
    """

    ## I expect to add more later, so users should explicitly choose Genova2016 for forward compatibility. Mittelholz might be publishing hers soon.
    info = {
        'Langlais2019': {
            'metadata': {
                'description': 'Martian magnetic field model, based on 14386 ESD, inversion using MGS MAG, MGS ER and MAVEN MAG, field predicted at 150km altitude. SH model, ref surface =3393.5 km, from Langlais/Thebault/Houliez/Purucker/Lillis internal coefficients.',
                'units'      : 'nT',
                'lmax'       : lmax,
                'links'      : {
                    'data' : 'https://doi.org/10.5281/zenodo.3876714',
                    'paper': 'https://doi.org/10.1029/2018JE005854',
                },
            },
        },
    }

    if model not in info:
        raise ValueError(f"Invalid magnetic field model: '{model}'. Options are: {list(info.keys())}.")



    if model == 'Langlais2019':

        fpath = _get_fpath_dataset(model)

        '''
        - Arguments for `from_file` are taken directly from the `pyshtools` source code (`pyshools.datasets.Mars.Langlais2019()`).
            - I rewrite it so the dataset is downloaded to `redplanet` cache rather than `pyshtools` cache, ensuring `redplanet` can fully manage/clear its own dataset cache.
        '''
        ds = (
            pysh.shclasses.SHMagCoeffs.from_file(
                fpath,
                lmax       = lmax,
                skip       = 4,
                r0         = 3393.5e3,
                header     = False,
                file_units = 'nT',
                units      = 'nT',
                encoding   = 'utf-8',
            )
            .expand()
            .to_xarray()
            .isel(lat=slice(None, None, -1))  ## in pysh, lats are always decreasing at first
        )


        data_dict = {}
        for data_var in list(ds.data_vars):
            data_dict[data_var] = ds[data_var].values

        metadata = info[model]['metadata']
        metadata['fpath'] = fpath


        global _dat_mag
        _dat_mag = GriddedData(
            lon       = ds.lon.values,
            lat       = ds.lat.values,
            is_slon   = False,
            data_dict = data_dict,
            metadata  = metadata,
        )



    else:
        raise ValueError(f"THE DEVELOPER MESSED UP. THIS SHOULD NOT HAPPEN.")

    return



from collections.abc import Callable

from redplanet.helper_functions.docstrings.direct_subs.main import _substitute_direct
from redplanet.helper_functions.docstrings.references.main import _substitute_references


def substitute_docstrings(func: Callable) -> Callable:
    """
    Substitute common docstrings into a function's docstring.

    Parameters
    ----------
    func : Callable
        The function to modify.

    Returns
    -------
    Callable
        The modified function with common docstrings substituted.
    """

    if not func.__doc__:
        return func

    func = _substitute_direct(func)
    func = _substitute_references(func)

    return func



from collections.abc import Callable
import re
from textwrap import indent
from importlib.resources import files

from redplanet.helper_functions.docstrings.references.Bibtex import BibtexDatabase


__bibtex_db: BibtexDatabase = BibtexDatabase(
    fpath = files('redplanet.helper_functions.docstrings.references') / 'references.bib'
)

## match {@key} anywhere in the docstring (even mid-sentence), capturing preceding whitespace (if any)
__pattern = re.compile(r'\{@([^}]+)\}')


def __cite_intext(key: str) -> list[str, str]:
    """
    Format an intext citation

    Parameters
    ----------
    key : str
        Format is "ref.f", where "ref" references the bib file, and "f" is the format of intext citation which can be "p" (parenthetical) or "n" (textual).

    Returns
    -------
    list[str, str]
        Formatted intext and full citations respectively.

    Raises
    ------
    ValueError
        If the `key` is not in the correct format (see parameter definition).
    KeyError
        If the "ref" part of the `key` is not in the bib file.
    """

    parts = key.rsplit('.', 1)

    if len(parts) == 1:
        raise ValueError(f'Invalid intext citation: "{key}" -- Must be in the form "key.f", where "key" references the bib file, and "f" is the format of intext citation which can be "p" (parenthetical) or "n" (textual).')

    ref_key, citation_format = parts

    ## note: errors for `ref_key` not in bib file or an invalid `citation_format` are handled by `BibtexDatabase.cite()`
    return [
        __bibtex_db.cite(ref_key, citation_format),
        __bibtex_db.cite(ref_key, 'full')
    ]



def _substitute_references(func: Callable) -> Callable:

    ## 1/2: substitute the intext citations.
    full_citations = set()

    def repl(match: re.Match) -> str:
        ref_key = match.group(1)
        intext, full = __cite_intext(ref_key)
        full_citations.add(full)
        return intext

    func.__doc__ = __pattern.sub(repl, func.__doc__)

    if not full_citations:
        return func

    ## 2/2: add the full citations to the end of the docstring.
    full_citations = sorted(list(full_citations))

    refs = [
        'References',
        '----------',
    ]
    refs.extend(f'{i}. {cite}' for i, cite in enumerate(full_citations, 1))
    refs = '\n'.join(refs)

    ## add indentation to match the rest of the docstring, which is taken from the first non-empty line
    for line in func.__doc__.split('\n'):
        if line.strip():
            indentation = line[:len(line) - len(line.lstrip())]
            refs = indent(refs, indentation)
            break

    func.__doc__ += f'\n\n{refs}'

    return func



import numpy as np

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

from redplanet import Crust





lon_bounds = [-180, 180]
lat_bounds = [-90 , 90]

lons = np.linspace(lon_bounds[0], lon_bounds[1], 1000)
lats = np.linspace(lat_bounds[0], lat_bounds[1], 1000)

Crust.topo.load('DEM_463m')
dat_topo = Crust.topo.get(lons, lats)





'''template for plotting stuff:'''
fig, ax = plt.subplots(figsize=(10, 10))

# ax.set_title('Bouguer Gravity Anomaly')
ax.set_xlabel('Longitude')
ax.set_ylabel('Latitude')

im = ax.imshow(
    dat_topo / 1.e3,
    cmap   = 'RdBu_r',
    # cmap   = 'Greys',
    origin = 'lower',
    aspect = 'auto',  ## this makes pixels into squares; alternative is 'equal'
    extent = [lon_bounds[0], lon_bounds[1], lat_bounds[0], lat_bounds[1]],
)

divider = make_axes_locatable(ax)
cax = divider.append_axes("right", size="3%", pad=0.2)  ## `size` sets colorbar width to X% of main axes; `pad` sets separation between axes and colorbar to X inches
cbar = fig.colorbar(im, cax=cax)
cbar.set_label('Topography (km)')

plt.show()



_dict_fragments = {
    'param.lon':
        '''
        lon : float | np.ndarray
            Longitude coordinate(s) in range [-180, 360].
        ''',
    'param.lat':
        '''
        lat : float | np.ndarray
            Latitude coordinate(s) in range [-90, 90].
        ''',
    'param.as_xarray':
        '''
        as_xarray : bool, optional
            If True, return the data as an `xarray.DataArray`. Default is False.
        ''',
    'return.GriddedData':
        '''
        float | np.ndarray | xr.DataArray
            Data values at the the input coordinates. The return type is determined as follows:

            - float: if both `lon` and `lat` are floats.
            - numpy.ndarray (1D): if one of `lon` or `lat` is a numpy 1D array and the other is a float.
            - numpy.ndarray (2D): if both `lon` and `lat` are numpy 1D arrays. The first dimension of output array corresponds to `lat` values.
            - xarray.DataArray: see `as_xarray` parameter (this takes precedence over the above types).
        ''',
    'fulldoc.get_dataset_GriddedData':
        '''
        Get the underlying dataset which is currently loaded.

        Returns
        -------
        GriddedData
            Instance of RedPlanet's `GriddedData` class, which stores all coordinate/data/metadata information and accessing/plotting methods.

        Raises
        ------
        ValueError
            If the dataset has not been loaded yet (see the `load` function for this module).

        See Also
        --------
        `redplanet.helper_functions.GriddedData`
        ''',
    'fulldoc.get_metadata':
        '''
        Get metadata for the dataset which is currently loaded.

        Returns
        -------
        dict
            Contains information about the dataset such as description, units, references, download links, local file path, etc.

        Raises
        ------
        ValueError
            If the dataset has not been loaded yet (see the `load` function for this module).
        ''',
    'note._load':
        '''
        NOTE: This method is private & less modular because there will only ever be one GRS dataset, so lazy loading upon the first access is fine. In contrast, in other modules like `Crust.topo` / `Crust.moho`, we want the user to explicitly/deliberately call `load(<model_params>)` so they're aware of different models and which one they're choosing.
        ''',
}



import numpy as np
import xarray as xr

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.coordinates import (
    _verify_coords,
    _slon2plon,
)
from redplanet.helper_functions.docstrings.main import substitute_docstrings



@substitute_docstrings
def is_above(
    lon       : float | np.ndarray,
    lat       : float | np.ndarray,
    as_xarray : bool = False,
) -> bool | np.ndarray | xr.DataArray:
    """
    Determine if the given point(s) are above the dichotomy boundary.

    See `help(redplanet.Crust.dichotomy.get_coords)` for the source of the dichotomy boundary coordinates data.

    Parameters
    ----------
    {param.lon}
    {param.lat}
    {param.as_xarray}

    Returns
    -------
    bool | np.ndarray | xr.DataArray
        Boolean array indicating whether the input coordinates are above the dichotomy boundary. If both inputs are 1D numpy arrays then it returns a 2D numpy array with shape `(len(lat), len(lon))`.
    """

    ## input validation
    _verify_coords(lon, lat)
    lon = _slon2plon(lon)

    lon = np.atleast_1d(lon)
    lat = np.atleast_1d(lat)

    ## load dataset
    dat_dichotomy_coords = get_coords()

    ## for each input longitude, find nearest dichotomy coordinates
    i_lons = np.searchsorted(dat_dichotomy_coords[:,0], lon, side='right') - 1
    llons, llats = dat_dichotomy_coords[i_lons].T
    rlons, rlats = dat_dichotomy_coords[i_lons+1].T

    ## linear interpolate between two nearest dichotomy coordinates to find threshold latitude
    tlats = llats + (rlats - llats) * ( (lon - llons) / (rlons - llons) )

    ## compare shape(y,1) with shape(x), which broadcasts to shape(y,x) with element-wise comparison
    result = lat[:, None] >= tlats

    ## remove singleton arrays/dimensions (i.e. one or both inputs were scalars)
    result = np.squeeze(result)
    if result.size == 1:
        return result.item()

    elif as_xarray:
        result = xr.DataArray(
            result,
            dims   = ("lat", "lon"),
            coords = {"lat": lat, "lon": lon},
        )
        # result = result.sortby('lat').sortby('lon')
    return result



@substitute_docstrings
def get_coords() -> np.ndarray:
    """
    Get a list of dichotomy boundary coordinates.

    The origin of the dataset is not fully clear. We use the file "dichotomy_coordinates-JAH-0-360.txt" (0.046 MiB) downloaded from {@Wieczorek2022_icta.n}. They attribute the data to *"Andrews-Hanna et al. (2008)"* which is ambiguous (to me atleast, I could be missing something obvious) — my best guess is {@dichotomy_paper.n}.

    Returns
    -------
    np.ndarray
        A numpy array of shape `(n, 2)` where `n` is the number of coordinates, and the two columns are longitude (0->360) and latitude respectively.
    """
    fpath = _get_fpath_dataset('dichotomy_coords')
    dat_dichotomy_coords = np.loadtxt(fpath)
    return dat_dichotomy_coords



import zipfile

import pandas as pd
import xarray as xr

from redplanet.DatasetManager.main import _get_fpath_dataset
from redplanet.helper_functions.GriddedData import GriddedData

from redplanet.helper_functions.docstrings.main import substitute_docstrings





_dat_grs: GriddedData | None = None

@substitute_docstrings
def get_dataset() -> GriddedData:
    """
    {fulldoc.get_dataset_GriddedData}
    """
    if _dat_grs is None:
        _load()
    return _dat_grs

@substitute_docstrings
def get_metadata() -> dict:
    """
    {fulldoc.get_metadata}
    """
    return dict(get_dataset().metadata)





@substitute_docstrings
def _load() -> None:
    """
    Load the GRS dataset.

    For more information on the dataset, see `help(redplanet.GRS.get)`.

    {note._load}
    """

    fpath = _get_fpath_dataset('GRS_v2')

    df = pd.read_excel(
        fpath,
        engine='calamine',  # `python-calamine` takes ~0.09s, while `openpyxl` takes ~0.6s! Old method (unzipping) takes ~0.1-0.2s altogether.
        na_values=0,
    ).dropna(axis=1, how='all')


    df.columns = (
        df.columns
        .str.lower()
        .str.replace('latitude', 'lat')
        .str.replace('longitude', 'lon')
        .str.replace(' wt%', '_concentration')
        .str.replace(' wt %', '_concentration')
        .str.replace(' ppm', '_concentration')
        .str.replace(' sigma', '_sigma')
    )

    for col in df.columns:
        if col in ['lat', 'lon']:
            continue

        if 'th' in col:
            scale_factor = 0.000001  # convert "ppm" to mass fraction (out of 1)
        else:
            scale_factor = 0.01      # convert "weight percent" (0-100) to mass fraction (out of 1)
        df[col] = df[col] * scale_factor

    df = df.set_index(['lat', 'lon'])


    ds = xr.Dataset.from_dataframe(df)
    ds['cl+h2o+s_concentration'] = ds['cl_concentration'] + ds['h2o_concentration'] + ds['s_concentration']
    ds['cl+h2o+s_sigma'] = ds['cl_sigma'] + ds['h2o_sigma'] + ds['s_sigma']

    # ds['cl+h2o+s_concentration'] = ds['cl+h2o+s_concentration'].fillna(0)
    # ds['cl+h2o+s_sigma'] = ds['cl+h2o+s_sigma'].fillna(0)


    data_dict = {}
    for var in ds.data_vars:
        data_dict[var] = ds[var].values

    global _dat_grs
    _dat_grs = GriddedData(
        lon       = ds.lon.values,
        lat       = ds.lat.values,
        is_slon   = True,
        data_dict = data_dict,
        metadata  = {
            'description'   : '2001 Mars Odyssey Gamma Ray Spectrometer Element Concentration Maps',
            'units'         : 'mass fraction (out of 1)',
            'grid_spacing'  : 5,
            'links'         : {
                'paper'     : 'https://doi.org/10.1029/2022GL099235',
                'download'  : 'https://data.mendeley.com/datasets/3jd9whd78m/1',
            },
            'fpath'         : fpath,
        },
    )

    return



import pytest
import numpy as np

from redplanet import Craters



def test_getall():
    assert Craters.get().shape == (2072, 22)


def test_getall_with_params():
    df = Craters.get(
        lon      = [-180,180],
        lat      = [-90,90],
        diameter = [0,9999],
    )
    assert df.shape[0] == 2072


def test_getall_with_params_plon():
    df = Craters.get(
        lon      = [0,360],
        lat      = [-90,90],
        diameter = [0,9999],
    )
    assert df.shape[0] == 2072


def test_get_filtered():
    df = Craters.get(
        lon      = [-60,60],
        lat      = [-30,30],
        diameter = [100,200],
    )
    assert df.shape[0] == 64


def test_get_aged():
    df = Craters.get(has_age=True)
    assert df.shape[0] == 73


def test_get_named():
    df = Craters.get(name=['Copernicus', 'Henry'])
    assert df.shape[0] == 2

def test_get_asdict():
    d = Craters.get(
        name    = 'Henry',
        as_dict = True,
    )
    assert len(d) == 1
    assert d[0].get('id') == '10-0-003901'

    d = Craters.get(
        name    = ['Copernicus', 'Henry'],
        as_dict = True,
    )
    assert len(d) == 2
    assert d[0].get('id') == '04-1-001446'
    assert d[1].get('id') == '10-0-003901'



import pytest
import numpy as np

from redplanet import GRS
from redplanet.helper_functions.coordinates import CoordinateError

## `get` (public function) -- get GRS data
class Test__GRS_get:

    def test__GRS_get__valid(self):

        ## plain concentration
        assert np.isclose(
            GRS.get('th', 2.5, -7.5),
            0.572947204e-6
        )


        ## plain sigma
        assert np.isclose(
            GRS.get('th', 2.5, -7.5, quantity='sigma'),
            0.047707241e-6
        )


        ## nearest value (i.e. no interpolation)
        assert np.isclose(
            GRS.get('th', 2.5, -7.5),
            GRS.get('th', 2.51, -7.51),
        )


        ## array inputs
        expected_dat = np.array([
            [6.24730945e-07, 6.15593433e-07, 6.23700500e-07, 6.23700500e-07],
            [5.72947204e-07, 5.93936563e-07, 6.16395473e-07, 6.16395473e-07],
            [5.63852787e-07, 6.22758865e-07, 6.61521018e-07, 6.61521018e-07],
            [5.63852787e-07, 6.22758865e-07, 6.61521018e-07, 6.61521018e-07]
        ])
        expected_lons = np.array([  2.5,  7.5,  12.5,  12.5])
        expected_lats = np.array([ -2.5, -7.5, -12.5, -12.5])

        #### array input: list
        lons = [  4,  8,  12,  12.1 ]
        lats = [ -4, -8, -12, -12.1 ]
        actual_dat = GRS.get('th', lons, lats)
        assert np.allclose(actual_dat, expected_dat)

        #### array input: np.ndarray
        lons = np.array(lons)
        lats = np.array(lats)
        actual_dat = GRS.get('th', lons, lats)
        assert np.allclose(actual_dat, expected_dat)


        ## wraparound / slon<->plon conversion
        assert np.isclose(
            GRS.get('th', -180, 2.5),
            GRS.get('th',  180, 2.5),
        )
        assert np.isclose(
            GRS.get('th',   0, 2.5),
            GRS.get('th', 360, 2.5),
        )
        assert np.isclose(
            GRS.get('th', -90, 2.5),
            GRS.get('th', 270, 2.5),
        )


        ## normalize
        assert np.isclose(
            GRS.get('th', 2.5, -7.5, normalize=True),
            (
                GRS.get('th', 2.5, -7.5)
                / (1 - (   GRS.get('cl' , 2.5, -7.5)
                         + GRS.get('h2o', 2.5, -7.5)
                         + GRS.get('s'  , 2.5, -7.5) ))
            ),
        )



    def test__GRS_get__invalid(self):

            ## invalid element
            with pytest.raises(ValueError, match="not in list of supported elements"):
                GRS.get('invalid_element', 2.5, -7.5)

            ## can't normalize volatile elements
            for element in ['cl', 'h2o', 's']:
                with pytest.raises(ValueError, match="Can't normalize a volatile element"):
                    GRS.get(element, 2.5, -7.5, normalize=True)

            ## out-of-range coordinates
            with pytest.raises(CoordinateError, match="Longitude"):
                GRS.get('th', 361, 0)
            with pytest.raises(CoordinateError, match="Latitude"):
                GRS.get('th', 0, 91)



from importlib import import_module
from importlib.metadata import version
import pytest

# Dictionary declaring all dependency groups and their packages
dependencies = {
    'Dependencies: required': [
        'cartopy',
        'numpy',
        'pandas',
        'pyshtools',
        'python-calamine',
        'scipy',
        'xarray',
        'xxhash',
    ],
    'Dependencies (optional): "interactive"': [
        'jupyter',
        'matplotlib',
        'plotly',
    ],
    'Dependencies (optional): "generate-datasets"': [
        'dask',
        'rioxarray',
        'zarr',
    ],
}

def test__import_dependencies():
    for group_name, packages in dependencies.items():
        print(f'\n\t- {group_name}')
        for package in packages:
            try:
                imported_package = import_module(package)
                print(f'\t\t> {package} = {version(package)}')
            except ImportError:
                print(f'\t\t> {package} = NOT AVAILABLE')



__all__ = []


from redplanet.user_config.datacache import (
    get_dirpath_datacache,
    set_dirpath_datacache,
)
__all__.extend([
    'get_dirpath_datacache',
    'set_dirpath_datacache',
])


from redplanet.user_config.max_hash_size import (
    get_max_size_to_calculate_hash_GiB,
    set_max_size_to_calculate_hash_GiB,
)
__all__.extend([
    'get_max_size_to_calculate_hash_GiB',
    'set_max_size_to_calculate_hash_GiB',
])


from redplanet.user_config.stream_hash_check import (
    get_enable_stream_hash_check,
    set_enable_stream_hash_check,
)
__all__.extend([
    'get_enable_stream_hash_check',
    'set_enable_stream_hash_check',
])



import hashlib
from pathlib import Path
import requests

import pytest

from redplanet.DatasetManager.download import _download_file_from_url





## `_download_file_from_url` (internal function) -- download file from URL to local path
class Test__download_file_from_url:

    ## Valid input
    def test__download_file_from_url(self, tmp_path):
        known_url = 'https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html'
        known_sha256 = 'c19249281275b6c8e3e7ba90edbfbde1998c47f4a3e8171c9a9ed46458495e75'

        # Use tmp_path to create a temporary file path
        fpath = tmp_path / 'tempfile.html'

        _download_file_from_url(known_url, fpath)
        assert fpath.exists(), "The file was not downloaded. This may be due to a bad network connection."

        sha256_hash = hashlib.sha256()
        with open(fpath, 'rb') as f:
            for byte_block in iter(lambda: f.read(4096), b''):
                sha256_hash.update(byte_block)
        computed_hash = sha256_hash.hexdigest()

        assert computed_hash == known_sha256, f"Hash mismatch: {computed_hash} != {known_sha256}"


    ## Invalid input: `url` is invalid
    def test__download_file_from_url_invalid_url(self, tmp_path):
        invalid_url = 'http://this.url.does.not.exist/invalid.html'

        fpath = tmp_path / 'tempfile.html'

        with pytest.raises(requests.exceptions.RequestException):
            _download_file_from_url(invalid_url, fpath)



from redplanet.Crust.moho.loader import load, get_dataset, get_metadata
from redplanet.Crust.moho.getter import get
from redplanet.Crust.moho.consts import get_registry

__all__ = [
    'load',
    'get_dataset',
    'get_metadata',
    'get',
    'get_registry',
]



import pytest
import numpy as np

from redplanet import Crust
from redplanet.DatasetManager.dataset_info import MohoDatasetNotFoundError



def test_load_invalid():

    ## invalid interior (`ValueError`)
    with pytest.raises(ValueError, match='Unknown interior model'):
        Crust.moho.load(
            interior_model    = 'meow',
            insight_thickness = 40,
            rho_south         = 2900,
            rho_north         = 2900,
        )

    ## valid interior, but model not found -- 1/2: custom error `MohoDatasetNotFoundError`
    with pytest.raises(MohoDatasetNotFoundError):
        Crust.moho.load(
            interior_model    = 'Khan2022',
            insight_thickness = 0,
            rho_south         = 0,
            rho_north         = 0,
        )

    ## valid interior, but model not found -- 2/2: fail silently
    found = Crust.moho.load(
        interior_model    = 'Khan2022',
        insight_thickness = 0,
        rho_south         = 0,
        rho_north         = 0,
        fail_silently     = True,
    )
    assert (not found)



def test_registry():
    Crust.moho.get_registry().shape == (21894, 4)



def test_load_get_valid():
    ## verify that moho spherical harmonic coefficients can be downloaded, loaded, accessed (plain value and crustal thickness), and switched between models
    lons = np.linspace(-180, 360, 100)
    lats = np.linspace(-90, 90, 100)

    Crust.moho.load(
        interior_model = 'Khan2022',
        insight_thickness = 39,
        rho_south = 2700,
        rho_north = 2700,
    )
    assert np.allclose( Crust.moho.get(lons, lats).mean()               , 3335011.619244110 )
    assert np.allclose( Crust.moho.get(lons, lats, crthick=True).mean() , 51386.64291032735 )

    Crust.moho.load(
        interior_model = 'Khan2022',
        insight_thickness = 41,
        rho_south = 2900,
        rho_north = 2900,
    )
    assert np.allclose( Crust.moho.get(lons, lats).mean()               , 3327826.760446502 )
    assert np.allclose( Crust.moho.get(lons, lats, crthick=True).mean() , 58571.50170793515 )


    ## wraparound
    lons = np.linspace(-180, 0, 100)
    lats = np.array([-60, 0, 60])
    assert np.allclose(
        Crust.moho.get(lons    , lats),
        Crust.moho.get(lons+360, lats)
    )



__all__ = []


from redplanet.helper_functions import (
    coordinates,
    geodesy,
    GriddedData,
    misc,
)
__all__.extend([
    'coordinates',
    'geodesy',
    'GriddedData',
    'misc',
])



import numpy as np

def compute_pressure(
    diameter_km            : float,
    x_vals_km              : float | np.ndarray,
    y_vals_km              : float | np.ndarray,
    v_proj_km_s            : float = 10,
    rho_proj_kg_m3         : float = 2900,
    rho_crust_kg_m3        : float = 2900,
    transition_diameter_km : float = 7,
    compressibility        : float = 1.5,
    bulk_sound_speed_km_s  : float = 3.5,
    pressure_decay_const   : float = 1.87,
    return_params          : bool  = False,
) -> np.ndarray | tuple[np.ndarray, dict]:
    """
    Compute the maximum subsurface shock pressures from an impact using the Rankine-Hugoniot relations.


    Parameters
    ----------
    diameter_km : float
        Observed crater diameter in kilometers. This is the primary crater-dependent input.
    x_vals_km : float | np.ndarray
        Horizontal (parallel to surface) point(s) in kilometers where the pressure is to be computed.
    y_vals_km : float | np.ndarray
        Vertical (perpendicular to surface) point(s) in kilometers where the pressure is to be computed.
    v_proj_km_s : float, optional
        Projectile velocity in kilometers per second. Default is 10 km/s.
    rho_proj_kg_m3 : float, optional
        Projectile density in kilograms per cubic meter. Default is 2900 kg/m^3.
    rho_crust_kg_m3 : float, optional
        Target (crust) density in kilograms per cubic meter. Default is 2900 kg/m^3.
    transition_diameter_km : float, optional
        Transition diameter (in kilometers) from simple to complex crater morphology. Default is 7 km.
    compressibility : float, optional
        Compressibility coefficient of the target material (dimensionless). Default is 1.5.
    bulk_sound_speed_km_s : float, optional
        Bulk sound speed in the target material in kilometers per second. Default is 3.5 km/s.
    pressure_decay_const : float, optional
        Exponential decay constant for the pressure (or particle velocity) outside the isobaric core. Default is 1.87.
    return_params : bool, optional
        If True, the function will also return a dictionary of intermediate parameters. Default is False.


    Returns
    -------
    P_eff : np.ndarray
        Shock pressures in GPa at the specified point(s), with shape `(len(y_vals_km), len(x_vals_km))`.

    params : dict
        Only returned if `return_params` is True.

        A dictionary with the following keys:

        - `v_proj_km_s` — Projectile velocity (km/s).
        - `rho_proj_kg_m3` — Projectile density (kg/m^3).
        - `rho_crust_kg_m3` — Crust density (kg/m^3).
        - `transition_diameter_km` — Transition crater diameter (km).
        - `compressibility` — Compressibility coefficient.
        - `bulk_sound_speed_km_s` — Bulk sound speed (km/s).
        - `pressure_decay_const` — Pressure decay exponent.
        - `transient_diameter_km` — Transient crater diameter (km).
        - `E_proj_J` — Projectile kinetic energy (Joules).
        - `proj_radius_km` — Projectile radius (km).
        - `isobaric_radius_km` — Isobaric core radius (km).
        - `u_ic_km_s` — Particle velocity in the isobaric core (km/s).
        - `rise_time` — Shock pressure rise time (s).


    Notes
    -----
    For a full explanation/derivation of methods and equations, see "Supplemental > Impact Demagnetization" in the RedPlanet documentation website.
    """

    ## Convert inputs to SI and shorthands
    D_o       = diameter_km * 1e3
    x_1d      = x_vals_km * 1e3
    y_1d      = y_vals_km * 1e3
    v_proj    = v_proj_km_s * 1e3
    rho_proj  = rho_proj_kg_m3
    rho_crust = rho_crust_kg_m3
    D_star    = transition_diameter_km * 1e3
    C         = bulk_sound_speed_km_s * 1e3
    S         = compressibility
    n         = pressure_decay_const
    g         = 3.72076
    ## EVERYTHING FROM HERE FORWARD IS SI

    D_tr = 0.7576 * D_o**0.921 * D_star**0.079

    E_proj = ((1/0.2212) * D_tr * v_proj**0.09 * g**0.22)**(1/0.26)
    r_proj = ((3 * E_proj) / (2 * np.pi * rho_proj * v_proj**2))**(1/3)

    R_ic = r_proj * 0.7
    u_ic = 0.5 * v_proj
    tau_rise = r_proj / v_proj

    x_2d, y_2d = np.meshgrid(x_1d, y_1d)  ## X and Y are each 2D arrays, now calculations can be better broadcasted/vectorized

    R_dir = np.sqrt(x_2d**2 + (y_2d - R_ic)**2)
    R_ref = np.sqrt(x_2d**2 + (y_2d + R_ic)**2)
    delta_t = (R_ref - R_dir) / C

    def calc_u_p(r):
        return np.where(
            r <= R_ic,
            u_ic,
            u_ic*(r/R_ic)**(-n)
        )

    def calc_P_direct(r):
        u_p = calc_u_p(r)
        return rho_crust * u_p * (C + S*u_p)

    P_direct    = calc_P_direct(R_dir)
    P_reflected = calc_P_direct(R_ref)

    P_eff = np.where(
        delta_t >= tau_rise,
        P_direct,
        P_direct - P_reflected*(1 - delta_t/tau_rise)
    )

    P_eff = P_eff * 1e-9
    P_eff = np.squeeze(P_eff)

    if return_params:
        return (
            P_eff,
            {
                'v_proj_km_s'            : v_proj_km_s,
                'rho_proj_kg_m3'         : rho_proj_kg_m3,
                'rho_crust_kg_m3'        : rho_crust_kg_m3,
                'transition_diameter_km' : transition_diameter_km,
                'compressibility'        : compressibility,
                'bulk_sound_speed_km_s'  : bulk_sound_speed_km_s,
                'pressure_decay_const'   : pressure_decay_const,
                'transient_diameter_km'  : D_tr * 1e-3,
                'E_proj_J'               : E_proj,
                'proj_radius_km'         : r_proj * 1e-3,
                'isobaric_radius_km'     : R_ic * 1e-3,
                'u_ic_km_s'              : u_ic * 1e-3,
                'rise_time'              : tau_rise,
            }
        )
    return P_eff



import numpy as np
import pandas as pd

from redplanet.DatasetManager.main import _get_fpath_dataset

from redplanet.helper_functions.coordinates import _slon2plon
from redplanet.helper_functions.docstrings.main import substitute_docstrings



_df_craters: pd.DataFrame | None = None


def _get_dataset() -> pd.DataFrame:
    """
    Returns the full crater dataset. If the dataset has not been loaded yet, it will be loaded.

    Returns
    -------
    pd.DataFrame
        For description of columns, see `help(redplanet.Craters.get)`.
    """
    if _df_craters is None:
        _load()
    return _df_craters


@substitute_docstrings
def _load() -> None:
    """
    Load the crater dataset.

    For more information on the dataset, see `help(redplanet.Craters.get)`.

    {note._load}
    """
    fpath_df = _get_fpath_dataset('crater_db')
    global _df_craters
    _df_craters = pd.read_csv(fpath_df)

    _df_craters['plon'] = _slon2plon(_df_craters['lon'])
    _df_craters = _df_craters.replace(np.nan, None)

    ## convert ages (e.g. '4.00;-0.08;0.05') to a list of floats
    columns_to_convert = [col for col in _df_craters.columns if col.startswith('N_') or col.endswith('Age')]
    def convert(val):
        if isinstance(val, str):
            parts = val.split(';')
            if len(parts) == 3:
                try:
                    return [float(part) for part in parts]
                except ValueError:
                    return None
        return None
    for col in columns_to_convert:
        _df_craters[col] = _df_craters[col].apply(convert)

    return



import pytest

from redplanet.DatasetManager.dataset_info import (
    _get_download_info,
    DatasetNotFoundError,
    MohoDatasetNotFoundError,
)



## `_get_download_info` (internal function) -- returns information to download a dataset
class Test__get_download_info:

    ## Valid input
    def test__get_download_info__grs_fname(self):
        result = _get_download_info(name='dichotomy_coords')
        assert result['fname'] == 'dichotomy_coordinates-JAH-0-360.txt'

    ## Invalid input: `name` not available
    def test__get_download_info__invalid_dataset(self):
        with pytest.raises( DatasetNotFoundError ):
            _get_download_info(name='https://fauux.neocities.org/')



import numpy as np
import pandas as pd

from redplanet.Mag.depth.loader import get_dataset

from redplanet.helper_functions import geodesy
from redplanet.helper_functions.coordinates import _plon2slon





def get_nearest(
    lon     : float,
    lat     : float,
    as_dict : bool = False,
) -> pd.DataFrame | list[dict]:
    """
    Get magnetic source depth data, sorted from closest to furthest from the given point.

    For source of the dataset, see references of `help(redplanet.Mag.depth.get_nearest)`.

    Parameters
    ----------
    lon : float
        Longitude coordinate in range [-180, 360].
    lat : float
        Latitude coordinate in range [-90, 90].
    as_dict : bool, optional
        If True, return the data as a list of dictionaries. Default is False.

    Returns
    -------
    pd.DataFrame | list[dict]
        Information about all 412 dipoles, sorted from closest to furthest from the given input coordinate. Columns are identical to those in `redplanet.Mag.depth.get_dataset` (look there for full explanations), with the addition of a computed column:

        - `distance_km` : float
            - Distance from the given input coordinate to the dipole, in km.
    """

    lon = _plon2slon(lon)

    df_depths = get_dataset().copy()

    distances_km = geodesy.get_distance(
        start = np.array([lon, lat]),
        end   = df_depths[['lon', 'lat']].to_numpy(),
    )[:,0] / 1.e3

    df_depths['distance_km'] = distances_km
    df_depths.sort_values('distance_km', inplace=True)

    if as_dict:
        df_depths = df_depths.to_dict(orient='records')

    return df_depths



from collections.abc import Callable

import numpy as np

from redplanet.helper_functions import geodesy





def get_concentric_ring_coords(
    lon                 : float,
    lat                 : float,
    radius_km           : float,
    dist_btwn_rings_km  : float = 5,
    num_rings           : int   = None,
    dist_btwn_points_km : float = 5,
) -> tuple[ np.ndarray, tuple[np.ndarray] ]:
    """
    Generate concentric ring coordinates around a central point.

    This function computes the radii for a series of concentric rings centered at the specified location and then generates (longitude, latitude) coordinates for equally spaced points along each ring. The rings can be defined by either a fixed distance between rings or by specifying the total number of rings to generate.

    Parameters
    ----------
    lon : float
        Longitude coordinate of the center of the circle, in range [-180, 360].
    lat : float
        Latitude coordinate of the center of the circle, in range [-90, 90].
    radius_km : float
        Radius (in kilometers) of the largest/outermost ring.
    dist_btwn_rings_km : float, optional
        Distance (in kilometers) between consecutive rings. Must be specified if `num_rings` is not provided. Default is 5 km.
    num_rings : int, optional
        Total number of rings to generate. If this is specified then you must provide `dist_btwn_rings_km=None`.
    dist_btwn_points_km : float, optional
        Desired spacing (in kilometers) between adjacent points on each ring. Default is 5.

    Returns
    -------
    ring_radius_km__per_ring : np.ndarray
        Ring radii (in kilometers) for each ring.
    ring_coords__per_ring : tuple[np.ndarray]
        A tuple of 2D numpy arrays, each containing the (longitude, latitude) coordinates of points on the corresponding ring (shape is `(num_points,2)`).

    Raises
    ------
    ValueError
        If either both or neither of `dist_btwn_rings_km` and `num_rings` are specified.
    """

    if not (bool(dist_btwn_rings_km) ^ bool(num_rings)):  # XOR
        raise ValueError('Must specify either `dist_btwn_rings_km` or `num_rings`, not neither/both.')

    ## Get radii for a series of concentric rings, starting at the center and going up to a distance of `radius_km`.
    if dist_btwn_rings_km:
        ring_radius_km__per_ring = np.arange(0, radius_km, dist_btwn_rings_km)
    else:
        ring_radius_km__per_ring = np.linspace(0, radius_km, num_rings)

    ## Calculate the number of points that can fit in each ring such that each point is `dist_btwn_points_km` away from its neighbors.
    ## Given a circle with radius `r`, the number of points you could fit around the circumference (`2*pi*r`) such that every point is distance `x` from its neighbors is given by `2*pi*r/x`.
    num_points__per_ring = np.ceil(2 * np.pi * ring_radius_km__per_ring / dist_btwn_points_km).astype(int)
    ## note: we're using ceil here to ensure atleast 1 point per ring :)    (i think???)

    ## Enforce a minimum number of points per ring
    min_num_points = 10
    num_points__per_ring[num_points__per_ring < min_num_points] = min_num_points

    ## Generate (lon,lat) coordinates for each point on each ring.
    ring_coords__per_ring = []
    for (ring_radius_km, num_points) in zip(ring_radius_km__per_ring, num_points__per_ring):
        ## Generate the (lon,lat) coordinates of `num_points` points around the circle of radius `ring_radius_km`.
        ring_coords = geodesy.make_circle(
            lon       = lon,
            lat       = lat,
            radius    = ring_radius_km * 1e3,
            n_samples = num_points,
            endpoint  = False,
        )
        ring_coords__per_ring.append(ring_coords)

    return (
        ring_radius_km__per_ring,
        tuple(ring_coords__per_ring)
    )





def get_profile(
    ring_coords__per_ring : tuple[np.ndarray],
    accessor              : Callable,
    return_stats          : bool = False,
) -> np.ndarray | tuple[ np.ndarray, np.ndarray, tuple[np.ndarray] ]:
    """
    Compute a radial profile using data extracted from concentric rings.

    This function computes a one-dimensional radial profile by applying an accessor function to a set of coordinates along each ring. For each ring, it averages the values returned by the accessor function. Optionally, the function can also return additional statistical information.


    Parameters
    ----------
    ring_coords__per_ring : tuple[np.ndarray]
        A tuple where each element is a numpy array containing (longitude, latitude) coordinate pairs for a ring. This corresponds to the second output of `get_concentric_ring_coords`.
    accessor : Callable
        A function that accepts two arguments (longitude and latitude) and returns a numerical value corresponding to a data point at that location.
    return_stats : bool, optional
        If True, the function returns additional statistical data (standard deviation and raw values for each ring) along with the averaged values. Default is False.


    Returns
    -------
    avg_vals__per_ring : np.ndarray
        Averaged values per ring (starting with the smallest).

    sigma__per_ring : np.ndarray
        Only returned if `return_stats` is True.

        Standard deviations (sigma) per ring.

    vals__per_ring : tuple[np.ndarray]
        Only returned if `return_stats` is True.

        A tuple of 1D numpy arrays, each containing the data values extracted from each ring (shape is `(num_points,)`).
    """

    vals__per_ring = []
    for ring_coords in ring_coords__per_ring:
        vals = []
        for (lon, lat) in ring_coords:
            vals.append(accessor(lon, lat))
        vals = np.array(vals)
        vals__per_ring.append(vals)

    avg_vals__per_ring = np.array([np.mean(vals) for vals in vals__per_ring])

    if not return_stats:
        return avg_vals__per_ring

    sigma__per_ring = np.array([np.std(vals) for vals in vals__per_ring])

    return (
        avg_vals__per_ring,
        sigma__per_ring,
        tuple(vals__per_ring),
    )



from pathlib import Path

from redplanet.user_config import (
    get_dirpath_datacache,
    get_enable_stream_hash_check,
    get_max_size_to_calculate_hash_GiB,
)
from redplanet.DatasetManager.dataset_info import (
    _get_download_info,
    _get_download_info_moho,
)
from redplanet.DatasetManager.download import _download_file_from_url
from redplanet.DatasetManager.hash import (
    _calculate_hash_from_file,
    _calculate_hash_from_url,
    get_available_algorithms,
)



def _get_fpath_dataset(dataset_name: str) -> Path:
    """
    Get the file path of a dataset...
        1. The first time, we download it to a local cache folder (you can see the default location with `import redplanet; redplanet.get_dirpath_datacache()` or change with with `set_dirpath_datacache()`. Since downloading random files from the internet is risky (you're essentially trusting me with unrestricted access to your computer, not to mention a potential man-in-the-middle attack although unlikely) we take the following two safety measures:
            i. Before downloading, we verify the integrity of the file at the URL by calculating its hash "on the fly" (i.e. streaming it as opposed to fully downloading it) -- this ensures we don't download malicious/altered files, assuming you trust my intended file/hash.
            ii. After downloading, we again verify integrity of file on disk by calculating its hash, immediately deleting if it doesn't match. I can't think of any realistic case this would occur, but why not.
        2. Subsequent times, we see if a file with the correct name/hash already exists in the cache folder.

    Parameters:
    -----------
    dataset_name : str
        The name of the dataset to be retrieved.

    Returns:
    --------
    Path
        The file path to the cached or newly downloaded dataset.

    Raises:
    -------
    Exception
        If there is a hash mismatch for either the file already in cache or the streamed/downloaded file.
    """

    ## Get download info for dataset
    if dataset_name.startswith('Moho-Mars-'):
        info: dict = _get_download_info_moho(
            model_name          = dataset_name[len('Moho-Mars-'):],
            fpath_moho_registry = _get_fpath_dataset('moho_registry'),
        )
    else:
        info: dict = _get_download_info(dataset_name)

    fpath_dataset: Path = get_dirpath_datacache() / info['dirpath'] / info['fname']


    ## Out of hashes listed in metadata, pick the fastest (assuming `get_available_algorithms()` lists them from fastest to slowest)
    known_hash_value = None
    for hashalg in get_available_algorithms():
        known_hash_alg = hashalg
        known_hash_value = info['hash'].get(known_hash_alg)
        if known_hash_value is not None:
            break
    if (known_hash_value is None):
        raise ValueError(f"[Internal/unexpected error] Hash value not found for dataset '{dataset_name}'. Pester the developer.")


    ## Case 1: File not found in cache, download it.
    if (not fpath_dataset.is_file()):

        ## First, verify the integrity of the file at the URL by calculating its hash "on the fly" (i.e. streaming it as opposed to fully downloading it) -- this ensures we don't download malicious/altered files, assuming you trust my intended file/hash
        ## (Users can skip this with `redplanet.set_enable_stream_hash_check(False)`)
        if get_enable_stream_hash_check():
            calculated_hash_value_fromStream = _calculate_hash_from_url(info['url'], known_hash_alg)
            if (calculated_hash_value_fromStream != known_hash_value):
                error_msg = [
                    f"We need to download the dataset from the known URL [1], but the calculated hash of the file at that URL [2] doesn't match the known hash [3]:",
                    f"    > [1] Known URL: \t{info['url']}",
                    f"    > [2] Calculated hash: \t{known_hash_alg}-{calculated_hash_value_fromStream}",
                    f"    > [3] Known hash: \t{known_hash_alg}-{known_hash_value}",
                    f"To see all known information about the datasets, run `import redplanet; from pprint import pprint; pprint(redplanet.peek_datasets())`.",
                    f"=> DOWNLOAD ABORTED.",
                ]
                raise Exception('\n'.join(error_msg))

        ## Second, proceed to download the file to `fpath_dataset`
        _download_file_from_url(info['url'], fpath_dataset)

        ## (Optional) Third, just to be sure, verify integrity of the recently-downloaded file by calculating the hash. I can't think of any realistic case this would fail, but why not.
        max_size_to_calculate_hash_GiB = get_max_size_to_calculate_hash_GiB()
        file_size_GiB = fpath_dataset.stat().st_size / (2**30)

        if (max_size_to_calculate_hash_GiB == None) or (file_size_GiB <= max_size_to_calculate_hash_GiB):  ## calculate hash if either of these are true
            calculated_hash_value = _calculate_hash_from_file(fpath_dataset, known_hash_alg)
            if (calculated_hash_value != known_hash_value):
                fpath_dataset.unlink()  # Deletes the recently-downloaded file for safety
                error_msg = [
                    f"We already verified the integrity of the file at the URL [1] from its hash [2] -- however, after downloading it to disk [3] and recalculating the hash for safety [4], it doesn't match the known hash [5] for some unknown reason:",
                    f"    > [1] Known URL: \t{info['url']}",
                    f"    > [2] Calculated hash at URL: \t{known_hash_alg}-{calculated_hash_value_fromStream}",
                    f"    > [3] File path (now deleted): \t{fpath_dataset}",
                    f"    > [4] Calculated hash: \t{known_hash_alg}-{calculated_hash_value}",
                    f"    > [5] Known hash: \t{known_hash_alg}-{known_hash_value}",
                    f"**The downloaded file has been deleted for safety.**",
                    f"To see all known information about the datasets, run `import redplanet; from pprint import pprint; pprint(redplanet.peek_datasets())`.",
                    f"=> PROCESS ABORTED.",
                ]
                raise Exception('\n'.join(error_msg))
        else:
            # print(f'TEMP DEBUG: Skipped calculating hash for file since actual size ({file_size_GiB:.2f} GiB) exceeds the maximum size ({max_size_to_calculate_hash_GiB:.2f} GiB) in user config.')
            pass


    # Case 2: File already exists in cache, verify the hash
    else:

        ## Error case: calculated hash of file on disk does not match the known hash.
        max_size_to_calculate_hash_GiB = get_max_size_to_calculate_hash_GiB()
        file_size_GiB = fpath_dataset.stat().st_size / (2**30)

        if (max_size_to_calculate_hash_GiB == None) or (file_size_GiB <= max_size_to_calculate_hash_GiB):  ## calculate hash if either of these are true
            calculated_hash_value = _calculate_hash_from_file(fpath_dataset, known_hash_alg)
            if (calculated_hash_value != known_hash_value):
                error_msg = [
                    f"Dataset already exists in cache folder [1], but the calculated hash [2] doesn't match the known hash [3]:",
                    f"    > [1] File path: \t{fpath_dataset}",
                    f"    > [2] Calculated hash: \t{known_hash_alg}-{calculated_hash_value}",
                    f"    > [3] Known hash: \t{known_hash_alg}-{known_hash_value}",
                    f"This may occur because a user/process manually changed the dataset file, or the dataset was updated in `redplanet`. We suggest you delete or move the old dataset file and try again.",
                    f"To see all known information about the datasets, run `import redplanet; from pprint import pprint; pprint(redplanet.peek_datasets())`.",
                    f"=> PROCESS ABORTED.",
                ]
                raise Exception('\n'.join(error_msg))
        else:
            # print(f'TEMP DEBUG: Skipped calculating hash for file since actual size ({file_size_GiB:.2f} GiB) exceeds the maximum size ({max_size_to_calculate_hash_GiB:.2f} GiB) in user config.')
            pass



    return fpath_dataset



"""
PURPOSE: Verify that downloaded data files are as expected (i.e. the author isn't distributing malicious/altered files).
"""


_enable_stream_hash_check: bool = False
## TODO: before publishing, initialize this to True -- keeping it false rn for debugging/dev purposes


def get_enable_stream_hash_check() -> bool:
    """
    Get the current value of the flag that determines whether we verify the hash of a file at URL by streaming before fully downloading it.

    Returns
    -------
    bool
    """
    return _enable_stream_hash_check


def set_enable_stream_hash_check(value: bool) -> None:
    """
    Set the flag that determines whether we verify the hash of a file at URL by streaming before fully downloading it.

    Parameters
    ----------
    value : bool
    """
    global _enable_stream_hash_check
    _enable_stream_hash_check = value
    return



"""
PURPOSE: Save yourself time by not calculating hashes for large local files (i.e. you're certain you haven't altered them, nor are they out of date).
"""


_max_size_to_calculate_hash_GiB: float | None = 1.0
## TODO: before publishing, initialize this to `None` so all hashes are calculated by default, and the user can decide if they want to disable the security feature. The logic for this is already in place in `redplanet.DatasetManager.main._get_fpath_dataset()`, so you just need to change the default value here.


def get_max_size_to_calculate_hash_GiB() -> float | None:
    """
    Get the maximum size of a local file (in GiB) for which a hash will be calculated in order to verify its integrity. If the file is larger than this size, the hash will not be calculated.

    Returns
    -------
    float | None

    Notes
    -----
    IMPLEMENTATION NOTE: if this returns `None`, then all hashes are calculated by default (to see implementation logic, check `redplanet.DatasetManager.main._get_fpath_dataset()`).
    """
    return _max_size_to_calculate_hash_GiB


def set_max_size_to_calculate_hash_GiB(value: float) -> None:
    """
    Set the maximum size of a local file (in GiB) for which a hash will be calculated in order to verify its integrity. If the file is larger than this size, the hash will not be calculated.

    Parameters
    ----------
    value : float
    """
    global _max_size_to_calculate_hash_GiB
    _max_size_to_calculate_hash_GiB = value
    return



import pytest
import numpy as np

from redplanet import Crust
from redplanet.helper_functions.coordinates import CoordinateError


def test__Crust_dichotomy__valid():

    ## ensure data integrity
    dat_dichotomy_coords = Crust.dichotomy.get_coords()
    assert dat_dichotomy_coords.shape == (1467, 2)
    assert dat_dichotomy_coords.mean() == 95.79914792285959

    ## test the poles (sanity-check)
    lon = np.arange(-180,360,1)
    assert  Crust.dichotomy.is_above(lon,  90).all()
    assert ~Crust.dichotomy.is_above(lon, -90).all()

    ## test some points around the equator to make sure lons are correct
    assert  Crust.dichotomy.is_above(np.arange(-180, -80), 0).all()
    assert ~Crust.dichotomy.is_above(np.arange( -75, 120), 0).all()
    assert  Crust.dichotomy.is_above(np.arange( 130, 270), 0).all()
    assert ~Crust.dichotomy.is_above(np.arange( 285, 360), 0).all()


def test__Crust_dichotomy__invalid():

    ## out-of-range coordinates
    with pytest.raises(CoordinateError, match="Longitude"):
        Crust.dichotomy.is_above(-181, 0)
    with pytest.raises(CoordinateError, match="Longitude"):
        Crust.dichotomy.is_above(361, 0)
    with pytest.raises(CoordinateError, match="Latitude"):
        Crust.dichotomy.is_above(0, -91)
    with pytest.raises(CoordinateError, match="Latitude"):
        Crust.dichotomy.is_above(0, 91)



from contextlib import contextmanager
import time

import numpy as np



@contextmanager
def timer(message='', d=3):
    """
    A context manager that measures the time taken to execute the code block within it.

    Parameters
    ----------
    message : str, optional
        A message to display when the code block is executed, by default nothing.
    d : int, optional
        The number of decimal places to display in the elapsed time, by default 3.

    Examples
    --------
    >>> from redplanet.helper_functions.misc import timer
    >>> with timer('Elapsed time: '):
    ...     print(f'Sum of first 10^7 squared integers: {sum(x*x for x in range(10_000_000)):,}')

    Sum of first 10^7 squared integers: 333,333,283,333,335,000,000
    Elapsed time: 0.886 seconds
    """
    start = time.time()
    yield
    end = time.time()
    elapsed = end - start
    print(f"{message}{elapsed:.{d}f} seconds")



def find_closest_indices(
    sorted_array: np.ndarray,
    target_values: np.ndarray
) -> np.ndarray:
    """
    Find the indices of the closest elements in a sorted array for each target value.

    For each value in `target_values`, this function identifies the index of the element in `sorted_array` that is closest to it. In cases where a target value is exactly midway between two elements, the index of the left (smaller) element is returned.

    Parameters
    ----------
    sorted_array : np.ndarray
        A one-dimensional NumPy array sorted in ascending order. The array should contain numeric values.

    target_values : np.ndarray
        A one-dimensional NumPy array of target values for which to find the closest indices in `sorted_array`.

    Returns
    -------
    np.ndarray
        A one-dimensional NumPy array of integers, where each element is the index in `sorted_array` that is closest to the corresponding target value in `target_values`.


    Examples
    --------
    >>> import numpy as np
    >>> sorted_array = np.array([10, 20, 30, 40, 50])
    >>> target_values = np.array([25, 35, 5, 55])
    >>> find_closest_indices(sorted_array, target_values)
    array([1, 2, 0, 4])

    In this example:
    - For target `25`, the closest element is `20` at index `1`.
    - For target `35`, the closest element is `30` at index `2`.
    - For target `5`, the closest element is `10` at index `0`.
    - For target `55`, the closest element is `50` at index `4`.
    """
    insertion_indices = np.searchsorted(sorted_array, target_values)
    insertion_indices = np.clip(insertion_indices, 1, len(sorted_array) - 1)
    left_neighbors = sorted_array[insertion_indices - 1]
    right_neighbors = sorted_array[insertion_indices]
    closest_indices = np.where(
        np.abs(target_values - left_neighbors) <= np.abs(target_values - right_neighbors),
        insertion_indices - 1,
        insertion_indices
    )
    return closest_indices



import pytest

from redplanet import Crust



def test_substitute_docstrings():
    ## direct substitutions
    assert 'Get the underlying dataset' in Crust.topo.get_dataset.__doc__

    ## citations (read from bib file)
    assert 'Goddard Space Flight Center' in Crust.topo.load.__doc__

    ## multiple
    assert '{' not in Crust.topo.get.__doc__



from dataclasses import dataclass
from types import MappingProxyType
from pprint import pformat
from textwrap import dedent, indent

import numpy as np
import xarray as xr

from redplanet.helper_functions.misc import find_closest_indices
from redplanet.helper_functions.coordinates import (
    _verify_coords,
    _plon2slon,
    _slon2plon,
)
from redplanet.helper_functions.docstrings.main import substitute_docstrings





@dataclass(frozen=True)
class GriddedData:



    ## PUBLIC INSTANCE VARIALES -- The dataclass is frozen, meaning the attributes are immutable, so we can safely access `myobject.lon` directly.
    lon       : np.ndarray
    is_slon   : bool  ## True if the longitudes are "signed" (i.e. in the range [-180, 180]), False if they are "positive" (i.e. in the range [0, 360]).
    lat       : np.ndarray
    data_dict : dict[str, np.ndarray]  ## datasets (2D numpy arrays) must be indexed by `[lat, lon]`, corresponding to the order in `self.lat` and `self.lon` respectively.
    metadata  : dict

    @property
    def data_vars(self) -> list[str]:
        """
        Get the names/keys of the data variables in the dataset. These are potential values for the `var` parameter in the `get_values` method.
        """
        return list(self.data_dict.keys())





    def __post_init__(self):

        ## make lon/lat arrays immutable
        object.__setattr__(self, 'lon', self.lon.copy())
        object.__setattr__(self, 'lat', self.lat.copy())
        self.lon.flags.writeable = False
        self.lat.flags.writeable = False

        ## make dicts immutable
        object.__setattr__(self, 'data_dict', MappingProxyType(self.data_dict))
        object.__setattr__(self, 'metadata' , MappingProxyType(self.metadata))

        # ## make data arrays immutable -- NEVERMIND, this can be a `np.memmap` and I don't want to copy/mess with that
        # immutable_data_dict = {
        #     key: array.copy().view() for key, array in self.data_dict.items()
        # }
        # for array in immutable_data_dict.values():
        #     array.flags.writeable = False
        # object.__setattr__(self, 'data_dict', immutable_data_dict)

        return

    def __str__(self) -> str:
        l = []
        l.append(dedent(
            f'''\
            GriddedData object:

            - Data variables:
                - {self.data_vars}

            - Data shape:
                - num_lons = {len(self.lon)}  (spacing = {self.lon[1] - self.lon[0]})
                - num_lats = {len(self.lat)}  (spacing = {self.lat[1] - self.lat[0]})

            - Metadata:\
            '''
        ))
        l.append(indent(pformat(dict(self.metadata)), prefix='    '))
        l = '\n'.join(l)
        return l





    def to_dict(self) -> dict:
        return {
            'lon': self.lon,
            'lat': self.lat,
            'dat': self.data_dict,
            'metadata': self.metadata,
        }

    def to_xarray(self) -> xr.Dataset:
        """
        Convert the GriddedData object to an xarray Dataset with coordinates `(lat, lon)`, one or more data variables, and information/metadata about the dataset in the attributes.
        """
        ## TODO: return dask array if data is too large...? for now, don't expose this method for stuff like topo.
        dat_vars = {
            key: xr.DataArray(
                data = array,
                dims = ['lat', 'lon'],
                coords = {
                    'lat': self.lat,
                    'lon': self.lon,
                },
                attrs = {
                    'long_name': key,
                },
            )
            for key, array in self.data_dict.items()
        }
        return xr.Dataset(dat_vars, attrs=dict(self.metadata))




    @substitute_docstrings
    def get_values(
        self,
        lon                 : float | np.ndarray,
        lat                 : float | np.ndarray,
        var                 : str,
        as_xarray           : bool = False
    ) -> float | np.ndarray | xr.DataArray:
        """
        Get specified dataset values at the specified coordinates.

        Parameters
        ----------
        {param.lon}
        {param.lat}
        var : str
            The name of the data variable to extract (i.e. one value from `self.data_vars`).
        {param.as_xarray}

        Returns
        -------
        {return.GriddedData}

        Raises
        ------
        ValueError
            If `var` is not one of the available data variables in the dataset (i.e. one value from `self.data_vars`).

        Notes
        -----
        This indexing/accessing approach is a full order of magnitude (sometimes more) faster than accessing an xarray DataArray/Dataset when it comes to random point-like accesses (namely the coordinates in concentric rings when doing radial cross-sectioning averages).

        `GriddedDataset` is used in the following modules: `Crust.boug`, `Crust.moho`, `Crust.topo`, `GRS`, `Mag.sh`.
        """

        ## input validation
        if var not in self.data_vars:
            raise ValueError(f'Unknown data variable: "{var}".\nOptions are: {", ".join(self.data_vars)}.')

        _verify_coords(lon, lat)

        if self.is_slon:
            lon = _plon2slon(lon)
        else:
            lon = _slon2plon(lon)

        lon = np.atleast_1d(lon)
        lat = np.atleast_1d(lat)


        ## get data
        idx_lon = find_closest_indices(self.lon, lon)
        idx_lat = find_closest_indices(self.lat, lat)

        dat_full = self.data_dict[var]
        dat = dat_full[np.ix_(idx_lat, idx_lon)]

        dat = np.squeeze(dat)  ## remove singleton dimensions
        if dat.ndim == 0: dat = dat.item()


        if as_xarray:
            dat = xr.DataArray(
                data = dat,
                dims = ['lat', 'lon'],
                coords = {
                    'lat': lat,
                    'lon': lon,
                },
                attrs = dict(self.metadata),
            )

        return dat



import numpy as np
import xarray as xr

from redplanet.Crust.moho.loader import get_dataset

from redplanet.helper_functions.docstrings.main import substitute_docstrings



@substitute_docstrings
def get(
    lon       : float | np.ndarray,
    lat       : float | np.ndarray,
    crthick   : bool = False,
    as_xarray : bool = False
) -> float | np.ndarray | xr.DataArray:
    """
    Get Mohorovičić discontinuity depth (or derived crustal thickness) values at the specified coordinates. Dataset must be loaded first, see `redplanet.Crust.moho.load(...)`.

    Parameters
    ----------
    {param.lon}
    {param.lat}
    crthick : bool, optional
        If True, return crustal thickness values, which is just the difference between the moho and a spherical harmonic model of topography evaluated to the same degree (same method as {@Wieczorek2022_icta.n}). Default is False.
    {param.as_xarray}

    Returns
    -------
    {return.GriddedData}

        Units are meters [m].
    """

    dat_moho = get_dataset()

    return dat_moho.get_values(
        lon = lon,
        lat = lat,
        var = 'crthick' if crthick else 'moho',
        as_xarray = as_xarray,
    )



import pytest
import numpy as np

from redplanet import Mag


def test_load_invalid():
    ## verify that an invalid model name raises an error
    with pytest.raises(ValueError, match='Invalid'):
        Mag.sh.load('meow')


def test_load_get_valid():
    ## verify that magnetic field values can be downloaded, loaded, and accessed
    Mag.sh.load('Langlais2019')

    ### global mean
    lons = np.linspace(-180, 360, 100)
    lats = np.linspace(-90, 90, 100)
    assert np.allclose( Mag.sh.get(lons, lats).mean()                    , 455.1033973376897 )
    assert np.allclose( Mag.sh.get(lons, lats, quantity='radial').mean() , 16.08461829256901 )
    assert np.allclose( Mag.sh.get(lons, lats, quantity='theta' ).mean() , 1.340587063287876 )
    assert np.allclose( Mag.sh.get(lons, lats, quantity='phi'   ).mean() , 4.879693707789295 )

    ### wraparound
    lons = np.linspace(-180, 0, 100)
    lats = np.array([-60, 0, 60])
    assert np.allclose(
        Mag.sh.get(lons    , lats),
        Mag.sh.get(lons+360, lats)
    )



import pytest
import numpy as np

from redplanet import Crust


def test_load_invalid():
    ## verify that an invalid model name raises an error
    with pytest.raises(ValueError, match='Invalid'):
        Crust.boug.load('meow')


def test_load_get_valid():
    ## verify that bouguer data can be downloaded, loaded, and accessed
    Crust.boug.load('Genova2016')

    ### global mean
    lons = np.linspace(-180, 360, 100)
    lats = np.linspace(-90, 90, 100)
    assert np.allclose( Crust.boug.get(lons, lats).mean() , -18.574545132436604 )

    ### wraparound
    lons = np.linspace(-180, 0, 100)
    lats = np.array([-60, 0, 60])
    assert np.allclose(
        Crust.boug.get(lons    , lats),
        Crust.boug.get(lons+360, lats)
    )



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "box-sdk-gen[jwt]",
#     "pandas",
#     "tqdm",
# ]
# ///


from typing import List, Callable
from datetime import datetime
from pathlib import Path, PurePosixPath
import logging
from pprint import pp

from tqdm import tqdm
import pandas as pd

from box_sdk_gen import JWTConfig, BoxJWTAuth, BoxClient # box authentication
from box_sdk_gen import AddShareLinkToFileSharedLink, UpdateSharedLinkOnFileSharedLink, AddShareLinkToFileSharedLinkPermissionsField, UpdateSharedLinkOnFileSharedLinkPermissionsField # box link sharing (lol)



''' ######################################################################## '''
'''                                   MAIN                                   '''
''' ######################################################################## '''

def main():

    ''' ———————————————————————————— User inputs ——————————————————————————— '''

    root_folder_id : str = '270589643170'   # path: 'rubox:shared_permalinks/redplanet/redplanet_cache/' (https://rutgers.app.box.com/folder/270589643170)

    # fpath_jwt_config : Path = list((Path.cwd() / '.secret/').glob('*.json'))[0]
    fpath_jwt_config : Path = Path('.secret') / '1190807008_sblhtn5v_config.json'   # token from personal account:"iltcc", app:"redplanet_export_v4"



    ''' —————————————————————————————— Logging ————————————————————————————— '''

    log        : logging.Logger
    fpath_logs : Path
    log, fpath_logs = setup_logger()

    log.info(f'Logging to file: {fpath_logs}')



    ''' ———————————————————————————— Box Auth ———————————————————————————— '''

    client : BoxClient = authenticate_box(fpath_jwt_config)

    log.info('')
    log.info('Authenticated with Box API.')
    log.info('')
    log.info(f'[Root folder info:]')
    log.info(f"    - '{_get_folder_name(client, root_folder_id)}/' (id: {root_folder_id})")



    ''' ——————————————————————— Build Registry Part 1 —————————————————————— '''

    # func_exclude_str = "lambda path: path.name == 'shcoeffs'"

    # log.info('')
    # log.info(f'[Exclude function:]')
    # log.info(f'    - `{func_exclude_str}`')
    log.info('')
    log.info(f'...traversing...')

    registry_noshcoeffs : dict = add_fileinfo_to_registry(
        client,
        root_folder_id,
        # func_exclude = eval(func_exclude_str)
    )

    log.info('')
    log.info(f'[Built registry with # files:]')
    log.info(f'    - {len(registry_noshcoeffs)}')



    ''' ——————————————————————— Build Registry Part 2 —————————————————————— '''

    registry_noshcoeffs_dlurls, dict_fileshare_statuses = add_filedlurls_to_registry(client, registry_noshcoeffs, log)

    log.info('')
    log.info('')
    log.info('')
    log.info(f'[SUMMARY:]')
    log.info(f'    - Added share link: {dict_fileshare_statuses['num_added_share']}')
    log.info(f'    - Fixed share link: {dict_fileshare_statuses['num_fixed_share']}')
    log.info(f'    - Already shared:   {dict_fileshare_statuses['num_already_shared']}')



    ''' —————————————————————————————— Export —————————————————————————————— '''

    fpath_output = export_registry_to_csv(registry_noshcoeffs_dlurls)

    log.info('')
    log.info('')
    log.info(f'[EXPORTED REGISTRY TO CSV:]')
    log.info(f'    - {fpath_output = }')





''' ######################################################################## '''
'''                             HELPER FUNCTIONS                             '''
''' ######################################################################## '''


''' ———————————————————————————————— Logging ——————————————————————————————— '''

def _create_console_handler(
    min_level : int = logging.INFO,
    fmt       : str = '[%(asctime)s, %(levelname)s]\t%(message)s',
    datefmt   : str = '%I:%M:%S %p',
    # datefmt   : str = '%H:%M:%S',
) -> logging.Handler:

    console_handler = logging.StreamHandler()
    formatter       = logging.Formatter(fmt, datefmt)

    console_handler.setLevel(min_level)
    console_handler.setFormatter(formatter)

    return console_handler


def _create_file_handler(
    logfile   : str,
    min_level : int = logging.DEBUG,
    # fmt       : str = '[%(asctime)s, %(levelname)s]    %(message)s',
    fmt       : str = '[%(asctime)s, %(levelname)s] %(message)s',
    datefmt   : str = '%Y/%m/%d %I:%M:%S %p',
    # datefmt   : str = '%Y/%m/%d %H:%M:%S',
) -> logging.Handler:

    file_handler = logging.FileHandler(logfile)
    formatter    = logging.Formatter(fmt, datefmt)

    file_handler.setLevel(min_level)
    file_handler.setFormatter(formatter)

    return file_handler


def setup_logger(
    fpath_logs  : Path = None,
    logger_name : str  = 'user_logger'
) -> List:

    logger = logging.getLogger(logger_name)

    logger.setLevel(logging.DEBUG)

    if fpath_logs is None:
        dirpath_output = Path.cwd() / '.logs'
        dirpath_output.mkdir(exist_ok=True)
        fpath_logs = dirpath_output / f"redplanet_registry_{datetime.now().strftime('%Y-%m-%d_%H-%M')}.log"

    file_handler    : logging.Handler = _create_file_handler(fpath_logs)
    console_handler : logging.Handler = _create_console_handler()

    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    # Prevent the logger from propagating messages to the root logger
    logger.propagate = False


    return [logger, fpath_logs]


''' ——————————————————————————————— Box Auth ——————————————————————————————— '''

def authenticate_box(fpath_jwt_config : Path) -> BoxClient:
    jwt_config : JWTConfig  = JWTConfig.from_config_file(config_file_path=fpath_jwt_config)
    auth       : BoxJWTAuth = BoxJWTAuth(config=jwt_config)
    user_auth  : BoxJWTAuth = auth.with_user_subject(user_id="34633045408")   # impersonating user from account:"iliketochacha" ("Kasane Teto")
    client     : BoxClient  = BoxClient(auth=user_auth)
    return client


''' ————————————————————————— Box Folder Accessors ————————————————————————— '''

def _get_folder_dict(
    client             : BoxClient,
    folder_id          : str,
    entries_per_search : int = 100,
    offset             : int = 0,
) -> dict:
    """
    RETURN:
        `dict`, which looks like this:
            {'id': '270589643170',
            'etag': '1',
            'type': 'folder',
            'sequence_id': '1',
            'name': 'redplanet_cache',
            'created_at': '2024-06-18T00:30:18-07:00',
            'modified_at': '2024-06-22T23:58:19-07:00',
            'description': '',
            'size': 11322836466,
            'path_collection': {'total_count': 2,
                                'entries': [{'id': '0',
                                            'type': 'folder',
                                            'name': 'All Files'},
                                            {'id': '270582896601',
                                            'etag': '1',
                                            'type': 'folder',
                                            'sequence_id': '1',
                                            'name': 'redplanet'}]},
            'created_by': {'id': '18595935892',
                            'type': 'user',
                            'name': 'Zain Kamal',
                            'login': 'zk117@rutgers.edu'},
            'modified_by': {'id': '18595935892',
                            'type': 'user',
                            'name': 'Zain Kamal',
                            'login': 'zk117@rutgers.edu'},
            'content_created_at': '2024-06-18T00:30:18-07:00',
            'content_modified_at': '2024-06-22T23:58:19-07:00',
            'owned_by': {'id': '18595935892',
                        'type': 'user',
                        'name': 'Zain Kamal',
                        'login': 'zk117@rutgers.edu'},
            'parent': {'id': '270582896601',
                        'etag': '1',
                        'type': 'folder',
                        'sequence_id': '1',
                        'name': 'redplanet'},
            'item_status': 'active',
            'item_collection': {'limit': 100,
                                'total_count': 3,
                                'offset': 0,
                                'order': [{'by': 'type', 'direction': 'ASC'},
                                        {'by': 'name', 'direction': 'ASC'}],
                                'entries': [{'id': '271456434555',
                                            'etag': '0',
                                            'type': 'folder',
                                            'sequence_id': '0',
                                            'name': 'Crust'},
                                            {'id': '271455606507',
                                            'etag': '0',
                                            'type': 'folder',
                                            'sequence_id': '0',
                                            'name': 'GRS'},
                                            {'id': '271457846138',
                                            'etag': '0',
                                            'type': 'folder',
                                            'sequence_id': '0',
                                            'name': 'Mag'}]}}

    """
    folder_dict : dict = client.folders.get_folder_by_id(
        folder_id = folder_id,
        limit     = entries_per_search,
        offset    = offset,
    ).to_dict()
    return folder_dict


def _get_folder_name(client: BoxClient, folder_id: str) -> str:
    empty_folder_dict : dict = _get_folder_dict(
        client             = client,
        folder_id          = folder_id,
        entries_per_search = 0,
    )
    folder_name : str = empty_folder_dict['name']
    return folder_name


def _get_folder_entries(client: BoxClient, folder_id: str) -> List[dict]:
    """
    We can't just call `get_folder_dict(...)` and then access `['item_collection']['entries']`, since the API call only returns up to 1,000 entries at a time. Therefore we need to repeatedly access the folder while incrementing the `offset` argument.

    RETURN:
        `List[dict]`, where FOLDERS look like:
            {
                'id'         : '271456434555',
                'etag'       : '0',
                'type'       : 'folder',
                'sequence_id': '0',
                'name'       : 'Crust'
            }
        and FILES look like:
            {
                'id'          : '1568917463274',
                'etag'        : '1',
                'type'        : 'file',
                'sequence_id' : '1',
                'name'        : '._Moho-Mars-Khan2022-18-2550-2550.sh',
                'sha1'        : '1abc429954206064f4ceed2886e22e1ffb65703a',
                'file_version': {
                    'id'  : '1724084185674',
                    'type': 'file_version',
                    'sha1': '1abc429954206064f4ceed2886e22e1ffb65703a'
                },
            }

    """

    empty_folder_dict : dict = _get_folder_dict(
        client             = client,
        folder_id          = folder_id,
        entries_per_search = 0,
    )
    num_entries        : int        = empty_folder_dict['item_collection']['total_count']

    entries            : List[dict] = []
    entries_per_search : int        = 1_000

    for offset in range(0, num_entries, entries_per_search):
        this_folder_dict : dict = _get_folder_dict(
            client             = client,
            folder_id          = folder_id,
            entries_per_search = entries_per_search,
            offset             = offset,
        )
        this_entries : List[dict] = this_folder_dict['item_collection']['entries']
        entries += this_entries

    return entries



''' —————————————————————————— Box File Accessors —————————————————————————— '''

def _get_file_dict(client: BoxClient, file_id: str) -> dict:
    """
    RETURN:
        `dict`, which looks like this:
            {'id': '1568919222016',
            'etag': '1',
            'type': 'file',
            'sequence_id': '1',
            'name': '2022_Mars_Odyssey_GRS_Element_Concentration_Maps.zip',
            'sha1': '5aa4778a4600c5fb85df1c6e363bd429e9c9e096',
            'file_version': {'id': '1724085982816',
                            'type': 'file_version',
                            'sha1': '5aa4778a4600c5fb85df1c6e363bd429e9c9e096'},
            'description': '',
            'size': 208819,
            'path_collection': {'total_count': 4,
                                'entries': [{'id': '0',
                                            'type': 'folder',
                                            'name': 'All Files'},
                                            {'id': '270582896601',
                                            'etag': '1',
                                            'type': 'folder',
                                            'sequence_id': '1',
                                            'name': 'redplanet'},
                                            {'id': '270589643170',
                                            'etag': '1',
                                            'type': 'folder',
                                            'sequence_id': '1',
                                            'name': 'redplanet_cache'},
                                            {'id': '271455606507',
                                            'etag': '0',
                                            'type': 'folder',
                                            'sequence_id': '0',
                                            'name': 'GRS'}]},
            'created_at': '2024-06-22T21:59:11-07:00',
            'modified_at': '2024-06-22T21:59:11-07:00',
            'content_created_at': '2024-03-08T02:44:12-08:00',
            'content_modified_at': '2024-03-08T02:44:12-08:00',
            'created_by': {'id': '18595935892',
                            'type': 'user',
                            'name': 'Zain Kamal',
                            'login': 'zk117@rutgers.edu'},
            'modified_by': {'id': '18595935892',
                            'type': 'user',
                            'name': 'Zain Kamal',
                            'login': 'zk117@rutgers.edu'},
            'owned_by': {'id': '18595935892',
                        'type': 'user',
                        'name': 'Zain Kamal',
                        'login': 'zk117@rutgers.edu'},
            'shared_link': {'url': 'https://rutgers.box.com/s/i1dy31or67y030yhof3c39ts19emigzd',
                            'effective_access': 'open',
                            'effective_permission': 'can_download',
                            'is_password_enabled': False,
                            'download_count': 0,
                            'preview_count': 0,
                            'download_url': 'https://rutgers.box.com/shared/static/i1dy31or67y030yhof3c39ts19emigzd.zip',
                            'access': 'open',
                            'permissions': {'can_download': True,
                                            'can_preview': True,
                                            'can_edit': False}},
            'parent': {'id': '271455606507',
                        'etag': '0',
                        'type': 'folder',
                        'sequence_id': '0',
                        'name': 'GRS'},
            'item_status': 'active'}

    """
    file_dict : dict = client.files.get_file_by_id(file_id=file_id).to_dict()
    return file_dict


def _get_file_name(client: BoxClient, file_id: str) -> str:
    file_dict : dict = _get_file_dict(client=client, file_id=file_id)
    file_name : str = file_dict['name']
    return file_name




''' ——————————————————————————— Box File Sharing ——————————————————————————— '''

def _get_file_sharestatus(client: BoxClient, file_id: str) -> str:
    """
    RETURNS:
        One of the following lists:
            ['not shared']
            ['incorrect sharing permissions', file_shareinfo_dict]
            ['correct sharing permissions',   file_shareinfo_dict, download_url]
    """
    file_dict : dict = _get_file_dict(client=client, file_id=file_id)

    file_shareinfo_dict : dict = file_dict.get('shared_link')
    """
    file_shareinfo_dict looks like this:
        {'url'                : 'https://rutgers.box.com/s/i1dy31or67y030yhof3c39ts19emigzd',
        'effective_access'    : 'open',
        'effective_permission': 'can_download',
        'is_password_enabled' : False,
        'download_count'      : 0,
        'preview_count'       : 0,
        'download_url'        : 'https://rutgers.box.com/shared/static/i1dy31or67y030yhof3c39ts19emigzd.zip',
        'access'              : 'open',
        'permissions'         : {'can_download': True,
                                 'can_preview' : True,
                                 'can_edit'    : False}}
    """


    ''' ———————————————————————— CASE 1: Not shared ———————————————————————— '''
    if file_shareinfo_dict is None:
        return ['not shared']


    ''' —————————————————— CASE 2: Bad sharing permissions ————————————————— '''
    proper_perms = {
        'effective_access'    : 'open',
        'effective_permission': 'can_download',
        'is_password_enabled' : False,
        'access'              : 'open',
        'permissions'         :
            {
                'can_download': True,
                'can_preview' : True,
                'can_edit'    : False
            }
    }   # NOTE: only paid accounts can change time-until-unshare (optional key "unshared_at"), so those are undetected and must be resolved manually.

    def _is_subset_dict(subset_dict, main_dict):
        return all(item in main_dict.items() for item in subset_dict.items())

    if not _is_subset_dict(proper_perms, file_shareinfo_dict):
        return ['incorrect sharing permissions', file_shareinfo_dict]


    ''' ————————————————— CASE 3: Good sharing permissions ————————————————— '''
    return ['correct sharing permissions', file_shareinfo_dict, file_shareinfo_dict['download_url']]


def _add_file_dlurl(client: BoxClient, file_id: str) -> str:
    file_dict : dict = client.shared_links_files.add_share_link_to_file(
        file_id = file_id,
        fields = "shared_link",
        shared_link = AddShareLinkToFileSharedLink(
            access      = 'open', # AddShareLinkToFileSharedLinkAccessField.OPEN.value
            password    = None,
            unshared_at = None,
            permissions = AddShareLinkToFileSharedLinkPermissionsField(
                can_download = True,
                can_preview  = True,
                can_edit     = False,
            ),
        ),
    ).to_dict()
    file_download_url : str = file_dict['shared_link']['download_url']
    return file_download_url


def _update_file_dlurl(clinet: BoxClient, file_id: str) -> str:
    file_dict : dict = client.shared_links_files.update_shared_link_on_file(
        file_id = file_id,
        fields = "shared_link",
        shared_link = UpdateSharedLinkOnFileSharedLink(
            access      = 'open',
            password    = None,
            unshared_at = None,
            permissions = UpdateSharedLinkOnFileSharedLinkPermissionsField(
                can_download = True,
                can_preview  = True,
                can_edit     = False,
            ),
        ),
    ).to_dict()
    file_download_url : str = file_dict['shared_link']['download_url']
    return file_download_url



''' ————————————————————————— Registry Constructors ———————————————————————— '''


def add_fileinfo_to_registry(
    client         : BoxClient,
    root_folder_id : str,
    func_include   : Callable[[PurePosixPath], bool] = lambda x: True,
    func_exclude   : Callable[[PurePosixPath], bool] = lambda x: False,
    registry       : dict = None,
) -> dict:
    if registry is None:
        registry = {}

    def _traverse_folder(
        this_folder_id : str,
        this_registry  : dict,
        path_stack     : List[str] = [],
    ) -> dict:
        path_stack.append(_get_folder_name(client, this_folder_id))

        for entry in _get_folder_entries(client, this_folder_id): # List[dict]

            this_entry_fullpath : PurePosixPath = PurePosixPath(*path_stack, entry['name'])

            if (not func_include(this_entry_fullpath)) or (func_exclude(this_entry_fullpath)):
                continue

            match entry['type']:
                case 'folder':
                    _traverse_folder(
                        this_folder_id = entry['id'],
                        this_registry  = this_registry,
                        path_stack     = path_stack,
                    )
                case 'file':
                    this_registry[str(this_entry_fullpath)] = {
                        'box_id': entry['id'],
                        'sha1': entry['sha1'],
                    }

        path_stack.pop()
        return this_registry

    return _traverse_folder(root_folder_id, registry)



def add_filedlurls_to_registry(
    client       : BoxClient,
    registry_v1  : dict,
    log          : logging.Logger,
    progress_bar : bool = True,
) -> dict:
    """
    Assume that `registry_v1` is dict of sub-dicts, where sub-dicts look like:
        'redplanet_cache/Mag/Langlais2019.sh': {
            'box_id': '1568919057294',
            'sha1'       : '0625c76c9594d1bf22e3bfa6c17ca8ee36ac2d2a'
        }
    """
    registry_v2 : dict = {}

    idx_progress       : int = 1
    num_added_share    : int = 0
    num_fixed_share    : int = 0
    num_already_shared : int = 0

    iterable = tqdm(registry_v1.items(), desc="Adding download URLs to registry") if progress_bar else registry_v1.items()

    log.debug('')
    log.debug('')
    log.debug('')

    for filepath, fileinfo in iterable:

        file_id          : str  = fileinfo['box_id']

        progress_idx_str : str = f"{idx_progress:0{len(str(len(registry_v1)))}}/{len(registry_v1)}" # basically just "01/10", "02/10", ...
        log.debug(f'[{progress_idx_str}]')
        idx_progress += 1
        log.debug(f'    - [FILE INFO:]')
        log.debug(f'        - {filepath = }')
        # log.debug(f'        - {file_id = }')
        log.debug(f'        - {fileinfo = }')

        file_sharestatus : list = _get_file_sharestatus(client, file_id)
        file_dlurl       : str

        file_sharestatus_code = file_sharestatus[0]

        log.debug(f'    - [FILE SHARE STATUS:]')
        # log.debug(f'        - {file_sharestatus_code = }')
        log.debug(f'        - {file_sharestatus = }')

        match file_sharestatus[0]:
            case 'not shared':
                file_dlurl = _add_file_dlurl(client, file_id)
                num_added_share += 1

            case 'incorrect sharing permissions':
                file_dlurl = _update_file_dlurl(client, file_id)
                num_fixed_share += 1

            case 'correct sharing permissions':
                file_dlurl = file_sharestatus[2]
                num_already_shared += 1

        registry_v2[filepath] = fileinfo | {'download_url': file_dlurl}

        log.debug(f'        - {file_dlurl = }')

    return [registry_v2, {'num_added_share'    : num_added_share,
                          'num_fixed_share'    : num_fixed_share,
                          'num_already_shared' : num_already_shared,} ]


''' ———————————————————————————————— Export ———————————————————————————————— '''

def export_registry_to_csv(
    registry  : dict,
    fpath_csv : Path = None
) -> Path:
    if fpath_csv is None:
        dirpath_output = Path.cwd() / 'output'
        dirpath_output.mkdir(exist_ok=True)
        fpath_csv = dirpath_output / f"redplanet_registry_{datetime.now().strftime('%Y-%m-%d_%H-%M')}.csv"

    registry_list : List[dict] = [{'filepath': k, **v} for k, v in registry.items()]
    df = pd.DataFrame(registry_list)
    df.to_csv(fpath_csv, index=False)
    return fpath_csv


''' ######################################################################## '''
'''                                   Main                                   '''
''' ######################################################################## '''

if __name__ == '__main__':
    main()



from pathlib import Path
from importlib.resources import files
from dataclasses import dataclass, field





@dataclass(frozen=True)
class Name:
    raw: str

    @property
    def full(self) -> str:
        # e.g. "Hahn, Brian C." -> "Hahn, B. C."
        # e.g. "Goddard Space Flight Center" -> "Goddard Space Flight Center"
        if self._is_human:
            return f'{self.last}, {self._initials}'
        return self._cleaned

    @property
    def last(self) -> str:
        # e.g. "Hahn, Brian C." -> "Hahn"
        # e.g. "Goddard Space Flight Center" -> "Goddard Space Flight Center"
        if self._is_human:
            return self._cleaned.split(', ', 1)[0]
        return self._cleaned

    @property
    def _cleaned(self) -> str:
        return self.raw.strip()

    @property
    def _is_human(self) -> bool:
        # assume that human names will always have a comma, e.g. "Hahn, Brian C."
        return ',' in self._cleaned

    @property
    def _initials(self) -> str:
        # e.g. "Hahn, Brian C." -> "B. C."
        if self._is_human:
            # to get initials, split the first name(s) (e.g. "Brian C.") into tokens (e.g. ["Brian", "C."]) and abbreviate each one (e.g. ["B.", "C."])
            first = self._cleaned.split(', ', 1)[1]
            return ' '.join(f'{token[0]}.' for token in first.split())
        return ''





@dataclass(frozen=True)
class BibtexEntry:
    key: str
    entry_type: str
    fields: dict[str, str | list[str] | list[Name]]


    def cite(self, format: str) -> str:
        match format:

            case 'full' | 'f':

                match self.entry_type:
                    case 'article':
                        return self._cite_full_article()
                    case 'dataset' | 'misc':
                        return self._cite_full_dataset_or_misc()
                    case _:
                        raise ValueError(f'Function `cite()` received an invalid value for `entry_type` -- must be one of "article", "dataset", or "misc", but got: {self.entry_type}')

            case 'narrative' | 'n':
                return f'{self._authors_intext(parenthetical=False)} ({self.fields["year"]})'

            case 'parenthetical' | 'p':  # note: you must add the parentheses yourself
                return f'{self._authors_intext(parenthetical=True)}, {self.fields["year"]}'

            case _:
                raise ValueError(f'Function `cite()` received an invalid value for `format` "{format}" -- Must be one of "full"/"f", "narrative"/"n", or "parenthetical"/"p".')


    def _cite_full_article(self) -> str:
        """
        Build a full APA citation for an 'article'.
        """
        authors = self._authors_full()
        year    = self.fields['year']
        title   = self.fields['title']

        journal = self.fields['journal']
        volume  = self.fields.get('volume', '')
        issue   = self.fields.get('number', '')
        pages   = self.fields.get('pages' , '')

        link = self.fields.get('doi') or self.fields.get('url')
        if not link: raise ValueError(f'No DOI or URL found for entry {self.key}')
        link = self._format_link(link)

        note = self.fields.get('_note', '')
        if note:
            note = f' *[{note}]*'

        # build the journal info (italicized using asterisks)
        journal_info = f'*{journal}*'
        if volume:
            journal_info += f', {volume}'
        if issue:
            journal_info += f'({issue})'
        if pages:
            journal_info += f', {pages}'

        return f'{authors} ({year}). {title}. {journal_info}. {link}{note}'


    def _cite_full_dataset_or_misc(self) -> str:
        """
        Build a full APA citation for a 'dataset' (any type of dataset, on Zenodo or otherwise) or 'misc' (e.g. code base on Zenodo) type reference.
        """
        authors = self._authors_full()
        year    = self.fields['year']
        title   = self.fields['title']

        publisher = self.fields['publisher']  # data host, e.g. "USGS" or "Zenodo"

        link = self.fields.get('doi') or self.fields.get('url')
        if not link: raise ValueError(f'No DOI or URL found for entry {self.key}')
        link = self._format_link(link)

        note = self.fields.get('_note', '')
        if note:
            note = f' *[{note}]*'

        title = f'{title}'
        if self.entry_type == 'dataset':
            title += ' [Dataset]'

        return f'{authors} ({year}). *{title}*. {publisher}. {link}{note}'


    @staticmethod
    def _format_link(url: str) -> str:
        """
        formats a url so it's highlighted and opens in a new tab
        requires `mkdocs.yml` > `markdown_extensions:` to contain `attr_list`
        """
        return f'<{url}>' + '{target="_blank"}'


    def _authors_full(self) -> str:
        """
        Format a list of authors for a *full* APA citation.

        Input : [ Name(raw='Hahn, B. C.'), Name(raw='McLennan, S. M.'), Name(raw='Klein, E. C.') ]
        Output: "Hahn, B. C., McLennan, S. M., & Klein, E. C."
        """
        authors: list[Name] = self.fields['author']
        if len(authors) == 1:
            return authors[0].full
        else:
            return ', '.join(a.full for a in authors[:-1]) + f', & {authors[-1].full}'


    def _authors_intext(
        self,
        parenthetical: bool,
    ) -> str:
        """
        Format a list of authors for *narrative or parenthetical* APA citations.
            - For one author, use the last name.
            - For two authors, join last names with "and" (narrative) or "&" (parenthetical).
            - For three or more, use only the first author's last name plus "et al."

        Input : ['Gong, Shengxia', 'Wieczorek, Mark']
        Output: "Gong and Wieczorek" (parenthetical=False)
        Output: "Gong & Wieczorek"   (parenthetical=True)

        Input:  ['Hahn, Brian C.', 'McLennan, S. M.', 'Klein, E. C.']
        Output: "Hahn et al."
        """
        authors: list[Name] = self.fields['author']
        match len(authors):
            case 1:
                return authors[0].last
            case 2:
                join  = '&' if parenthetical else 'and'
                return f'{authors[0].last} {join} {authors[1].last}'
            case _:
                return f'{authors[0].last} et al.'





@dataclass(frozen=True)
class BibtexDatabase:
    fpath: Path
    entries: dict[str, BibtexEntry] = field(init=False)


    def __post_init__(self):
        bib_dict = self._read_bibtex_file(self.fpath)
        bib_dict = self._postprocess_clean_strings(bib_dict)
        bib_dict = self._postprocess_separate_list(bib_dict, 'keywords', ', ')  # I used to determine if something is a dataset if it's type 'misc' and 'dataset' was in the list of keywords, but that's hacky. Turns out `@dataset` is a valid BibTeX entry type -- that also frees up `@misc` to be used for entries in Zenodo that aren't datasets such as the SHTOOLS code.
        bib_dict = self._postprocess_separate_list(bib_dict, 'author', ' and ')
        bib_dict = self._postprocess_format_names(bib_dict)

        entries = {}

        for key in bib_dict:
            entry = BibtexEntry(
                key = key,
                entry_type = bib_dict[key]['type'],
                fields = bib_dict[key],
            )
            entries[key] = entry

        # self.entries = entries
        object.__setattr__(self, 'entries', entries)



    def cite(self, key: str, format: str) -> str:
        try:
            return self.entries[key].cite(format)
        except KeyError:
            raise KeyError(f'Key "{key}" not found in the BibTeX database. Options are: [{", ".join(self.keys)}]')



    @property
    def keys(self) -> set[str]:
        return set(self.entries.keys())



    @staticmethod
    def _read_bibtex_file(fpath: Path) -> dict[str, dict]:
        """
        Read a BibTeX file and convert to a dictionary with minimal processing.

        Sample input (contents of file):
            ```
            @misc{Goddard2003,
                author = {Goddard Space Flight Center},
                title = {{Mars MGS MOLA DEM 463m}},
                year = {2003},
            }
            ```

        Sample output (python dictionary):
            ```
            'Goddard2003': {
                'type': 'misc',
                'author': '{Goddard Space Flight Center}',
                'title': '{{Mars MGS MOLA DEM 463m}}',
                'year': '{2003}',
            }
            ```
        """
        lines = [
            line.strip()
            for line in
                fpath.read_text().split('\n')
        ]

        entries = {}
        current_key = None

        for i, line in enumerate(lines):

            # case 1: skip comments and non-fields (empty or closing brace)
            if line.startswith('%') or not line or line.startswith('}'):
                continue

            # case 2: new entry (e.g., '@article{...' )
            elif line.startswith('@'):
                this_type = line.split('{')[0][1:].lower()
                this_key = line.split('{')[1].split(',')[0].strip()
                if this_key in entries:
                    raise ValueError(f"Duplicate key: {this_key}")
                entries[this_key] = {'type': this_type}
                current_key = this_key
                continue

            # case 3: field
            else:
                if '=' not in line:
                    raise ValueError(f"Invalid line {i} (missing equals sign): {line}")
                    # continue
                if not line.endswith(','):
                    raise ValueError(f"Invalid line {i} (missing comma): {line}")
                    # continue

                (field_name, field_val) = line.split('=', 1)
                field_name = field_name.strip().lower()
                field_val = field_val.strip().rstrip(',')

                entries[current_key][field_name] = field_val

        return entries



    @staticmethod
    def _postprocess_clean_strings(d: dict[str, dict]) -> dict[str, dict]:
        """
        Convert all fields in a dictionary of BibTeX entries (see output of `_read_bibtex_file`) to plain strings. This basically means removing curly braces and quotes.

        Sample input:
            ```
            'Goddard2003': {
                'type': 'misc',
                'author': '{Goddard Space Flight Center}',
                'title': '{{Mars MGS MOLA DEM 463m}}',
                'year': '"2003"'
            }
            ```

        Sample output:
            ```
            'Goddard2003': {
                'type': 'misc',
                'author': 'Goddard Space Flight Center',
                'title': 'Mars MGS MOLA DEM 463m',
                'year': '2003'
            }
            ```
        """
        for key in d:
            for name, val in d[key].items():
                if name == 'type':
                    continue

                if val.startswith('{{') and val.endswith('}}'):
                    val = val[2:-2]
                elif (val.startswith('{') and val.endswith('}')) or (val.startswith('"') and val.endswith('"')):
                    val = val[1:-1]
                else:
                    raise ValueError(f"Invalid value: {val}")
                    # continue

                d[key][name] = val.strip()

        return d



    @staticmethod
    def _postprocess_separate_list(
        d          : dict[str, dict],
        field_name : str,
        sep        : str,
    ) -> dict[str, dict]:
        """
        In a dictionary of BibTeX entries (see output of `parse_bibtex_file`), convert a field from a single string to a list of strings.

        Input: "Hahn, B. C. and McLennan, S. M. and Klein, E. C."
        Output (sep=' and '): ['Hahn, B. C.', 'McLennan, S. M.', 'Klein, E. C.']

        Input: "dataset, MOLA, DEM"
        Output (sep=' '): ['dataset', 'MOLA', 'DEM']
        """
        for key in d:
            if field_name in d[key]:
                d[key][field_name] = d[key][field_name].split(sep)
        return d



    @staticmethod
    def _postprocess_format_names(d: dict[str, dict]) -> dict[str, dict]:
        """
        In a dictionary of BibTeX entries (see output of `parse_bibtex_file`), convert author names from list[str] to list[Name].
        """
        for key in d:
            if 'author' in d[key]:
                d[key]['author'] = [Name(name) for name in d[key]['author']]
        return d
